# fmt: off
# Disable formatter. This is deliberately formatted to stay within 80 characters
# per line. Black, however, formats some lines going beyond this.

"""Transformations for converting scaled dot-product attention patterns to FINN hardware layers.

This module provides transformations to detect and convert multi-node ONNX patterns
representing scaled dot-product attention into single FINN custom hardware operations,
enabling efficient FPGA implementation of transformer attention mechanisms.
"""

# Standard math functions
import math

# Need numpy for modifying the onnx graph tensors, which are numpy style arrays
import numpy as np

# Utility for handling ONNX nodes and tensors
from onnx import NodeProto
from onnx import helper as oh

# QONNX datatypes
from qonnx.core.datatype import BaseDataType, DataType

# QONNX wrapper of ONNX model graphs
from qonnx.core.modelwrapper import ModelWrapper

# Convert ONNX nodes to QONNX custom ops
from qonnx.custom_op.registry import getCustomOp

# QONNX graph transformation base class
from qonnx.transformation.base import Transformation

# Transformations running qonnx datatype inference
from qonnx.transformation.infer_datatypes import InferDataTypes

# Transformation running onnx shape inference
from qonnx.transformation.infer_shapes import InferShapes

# Gets items from protobuf by name
from qonnx.util.basic import get_by_name, remove_by_name

# Utility function for transforming ONNX graphs
from finn.transformation.util import (
    all_upstream_to_matmul,
    is_add,
    is_join_matmul,
    is_matmul,
    is_mul,
    is_softmax,
    op_types,
)

# Output warning messages
from finn.util.logging import log


class InferScaledDotProductAttention(Transformation):
    """Convert the operator pattern corresponding to scaled dot-product attention to
    the hardware custom operator node."""

    def apply(self, model: ModelWrapper):  # noqa
        """Apply the transform to a whole model graph."""
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # This transformation is triggered by finding a join-node MatMul
            if is_join_matmul(node, model):
                # If there are more than two branches feeding the MatMul, this
                # is probably not attention, softly skip the node
                if len(node.input) != 2:
                    continue
                # Follow both branches upstream looking for the next MatMul
                lhs, rhs = all_upstream_to_matmul(node, model)
                # Exactly one of the branches is supposed to contain a Softmax
                # operation
                if ("Softmax" in op_types(lhs)) == ("Softmax" in op_types(rhs)):
                    # TODO: Near match. But what is this? just skip?
                    continue
                # By convention and following the equation, the left hand side
                # of attention is the attention matrix, i.e., the one containing
                # Softmax and terminating in a join-node MatMul
                if "Softmax" not in op_types(lhs):
                    # Softmax must currently be on the right hand side, swap the
                    # order
                    lhs, rhs = rhs, lhs
                # The left hand side, i.e, attention matrix must terminate in a
                # join-node MatMul involving the query and key input
                if not is_join_matmul(lhs[-1], model):
                    # TODO: Near match. But what is this? just skip?
                    continue
                # Get shapes of input tensors, expect the second inputs, i.e.,
                # the keys to be transposed
                qh, ql, qe = model.get_tensor_shape(lhs[-1].input[0])
                kh, ke, kl = model.get_tensor_shape(lhs[-1].input[1])
                # The input shapes of the two matmul inputs must be compatible,
                # i.e., they must have matching embedding dimension
                if (qh, True, qe) != (kh, True, ke):
                    # Issue a warning of near match of the supported attention
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Mismatch in head or embedding dim at {lhs[-1].name}: "
                        f" {(qh, ql, qe)} vs. {(kh, kl, ke)}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue
                # There must be a Transpose feeding the key input
                transpose = model.find_producer(lhs[-1].input[1])
                # The transform applies only to transpose with exactly one input
                if transpose is None or len(transpose.input) != 1:
                    # Issue a warning of near match of the supported attention
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Missing Transpose near {lhs[-1].name}: "
                        f" {op_types([transpose])[0]}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # Skip this node if the transpose output forks into multiple
                # branches
                if model.is_fork_node(transpose):
                    # Issue a warning of near match of the supported attention
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Fork Transpose near {node.name}: {transpose.name}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # The input shape of the transpose must match the transpose
                # of the key matrix
                # @formatter:off
                assert model.get_tensor_shape(transpose.input[0]) == [
                    kh, kl, ke
                ]
                # @formatter:on
                # Collect the input tensors to the attention operation, i.e.,
                # the query, key and value tensors
                q, k, v = lhs[-1].input[0], transpose.input[0], None
                # Check all outputs of the right hand side branch for being the
                # values input (this could for example be a split feeding
                # multiple heads)
                for candidate in rhs[0].output:
                    if node in model.find_consumers(candidate):
                        v = candidate

                # Validate that the values are actually consumed by the final
                # matmul. For queries and keys this should all be given, as we
                # just walked upwards the graph.
                assert v is not None

                # Get the (optional) Softmax activation function
                act_a_softmax = lhs[0] if is_softmax(lhs[1]) else None
                # Get the (optional) query-key matmul activation function
                act_qk_matmul = lhs[-2] if is_matmul(lhs[-1]) else None

                # There might be no activation function between qk matmul and
                # softmax normalization
                if is_mul(act_qk_matmul) or is_softmax(act_qk_matmul):
                    # Remove the detected activation function node from the
                    # pattern candidates
                    act_qk_matmul = None

                def is_supported_activation(n: NodeProto):  # noqa: Shadows name
                    """Check whether the node is a supported type of activation.

                    Currently, only none-type and MultiThreshold activations are supported.
                    """
                    return n is None or n.op_type in {"MultiThreshold"}

                # Get the (optional) output matmul activation function
                act_av_matmul = model.find_direct_successors(node)
                # If the final matmul is a fork node, this needs to be handled
                # separately
                if act_av_matmul is not None and len(act_av_matmul) > 1:
                    # Assume no activation in this case
                    act_av_matmul = [None]
                # Unwrap the output activation from the list
                act_av_matmul, = act_av_matmul
                # The final activation can be omitted if it is not supported as
                # it might just be part of the next operator pattern
                if not is_supported_activation(act_av_matmul):
                    # Remove by setting to None (will be ignored by the next
                    # steps)
                    act_av_matmul = None
                # List all activations for validation and further processing
                #   Note: Order matters!
                acts = [act_qk_matmul, act_a_softmax, act_av_matmul]
                # Skip this node if any activation is not supported
                if not all(is_supported_activation(act) for act in acts):
                    # Issue a warning of near match of the supported attention
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Unsupported activation near {node.name}: "
                        f" One of {', '.join(op_types(acts))}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # Check whether there is a de-quantizer scale factor preceding
                # the Softmax operator
                dequant_softmax = lhs[2] if is_softmax(lhs[1]) else None

                # If there is no dequant softmax yet, check alternative pattern
                if dequant_softmax is None:
                    # Alternatively, there might not be a quantizer following
                    # the softmax
                    dequant_softmax = lhs[1] if is_softmax(lhs[0]) else None

                # Assume no attention mask by default
                mask, mask_mode, mask_dtype = [], 'none', DataType["BINARY"]
                # If there is an elementwise add operation where we have
                # expected the dequantizer, this might be an attention mask
                if is_add(dequant_softmax):
                    # Remember the candidate of the masking operation
                    maybe_mask = dequant_softmax
                    # If there is a mask candidate, the dequantizer, must be
                    # right before
                    dequant_softmax = model.find_direct_predecessors(
                        dequant_softmax
                    )
                    # The attention mask may not have multiple producers
                    if len(dequant_softmax) != 1:
                        # Issue a warning of near match of the supported
                        # attention pattern
                        # @formatter:off
                        log.warning(
                            f"{self.__class__.__name__}: Skipping near match: "
                            f"Unsupported de-quantizer near {maybe_mask.name}: "
                            f" {op_types(dequant_softmax)}"
                        )
                        # @formatter:on
                        # Skip transforming this instance
                        continue
                    # There is a single producer, which is probably the
                    # dequantizer
                    dequant_softmax, = dequant_softmax

                    # The mask can be an initializer or provided as an input. If
                    # it is given as an initializer, it can either be a causal
                    # mask or some arbitrary pattern.

                    def valid_mask(tensor):
                        """Check whether a tensor is a valid mask tensor.

                        Valid masks contain only two types of values: zero for not masked
                        and -inf for masked slots.
                        """
                        return all(
                            x in {0.0, -np.inf} for x in np.unique(tensor)
                        )

                    def is_causal(tensor):
                        """Check whether a tensor describes a causal attention mask.

                        Generates a causal mask of the same size and compares it against
                        the candidate tensor.
                        """
                        # Generate a causal mask of the same size
                        causal = np.triu(-np.inf * np.ones_like(tensor), 1)
                        # Compare candidate against the causal mask
                        return (tensor == causal).all()  # noqa: 'all'

                    # Try to get the initializer of the masking operation
                    mask_tensor = model.get_initializer(maybe_mask.input[1])
                    # Check whether this is constant mask known at export time
                    if mask_tensor is not None:
                        # We have a constant mask and need to validated that it
                        # only contains valid values
                        if not valid_mask(mask_tensor):
                            # Issue a warning of near match of the supported
                            # attention pattern
                            # @formatter:off
                            log.warning(
                                f"{self.__class__.__name__}: Skipping near"
                                f" match: Invalid values in mask near"
                                f" {maybe_mask.name}: {np.unique(mask_tensor)}"
                            )
                            # @formatter:on
                            # Skip transforming this instance
                            continue
                        # If this is a causal mask, just set the flag and drop
                        # the input as the behavior can be generated on the fly
                        if is_causal(mask_tensor):
                            # Set the mode flag
                            mask_mode = "causal"
                        # This is a constant but non-causal mask which needs to
                        # be kept as an input to the operator
                        else:
                            # Keep the input and set the mode flag
                            mask, mask_mode = [maybe_mask.input[1]], "const"
                            # Convert the mask to a binary mask getting rid of
                            # explicitly storing the infinities
                            mask_tensor = (mask_tensor == -np.inf)
                            # Set the initializer to the binary mask still using
                            # float as the container type
                            model.set_initializer(
                                *mask, mask_tensor.astype(np.float32)
                            )
                            # Set the quantization type annotation to binary
                            model.set_tensor_datatype(*mask, DataType["BINARY"])
                    # Dynamic input mask, cannot be validated beforehand
                    else:
                        # # Keep the input and set the corresponding mode flag
                        # mask, mask_mode = [maybe_mask.input[1]], "input"
                        # # Keep track of the datatype of the mask
                        # mask_dtype = model.get_tensor_datatype(*mask)

                        # Handling dynamic masks is more difficult and there is
                        # no solution for now.
                        # @formatter:off
                        log.warning(
                            f"{self.__class__.__name__}: Skipping near match: "
                            f"Unsupported dynamic mask near {maybe_mask.name}: "
                            f" {mask}"
                        )
                        # @formatter:on
                        # Skip transforming this instance
                        continue

                # Currently, only elementwise Mul is supported as de-quantizer
                if not is_mul(dequant_softmax):
                    # Issue a warning of near match of the supported attention
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f"Unsupported de-quantizer near {lhs[1].name}: "
                        f" {dequant_softmax.op_type}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # If there is a dequant scale factor, try to lift it from
                # initializer to node attribute
                if dequant_softmax is not None:
                    # Get the initializer tensor
                    scale = model.get_initializer(dequant_softmax.input[1])
                    # This must be an initializer, the attention operator
                    # currently does not handle any dynamically produced scale
                    # factors
                    if scale is None:
                        # Issue a warning of near match of the supported
                        # attention pattern
                        # @formatter:off
                        log.warning(
                            f"{self.__class__.__name__}: Skipping near match: "
                            f"Non-constant dequantizer near {node.name}: "
                            f" {dequant_softmax.name}"
                        )
                        # @formatter:on
                        # Skip transforming this instance
                        continue
                    # Currently, only scalar dequantizer scale factors are
                    # supported
                    if not all(x == 1 for x in scale.shape):
                        # Issue a warning of near match of the supported
                        # attention pattern
                        # @formatter:off
                        log.warning(
                            f"{self.__class__.__name__}: Skipping near match: "
                            f"Non-scalar dequantizer near {node.name}: "
                            f" {dequant_softmax.name}"
                        )
                        # @formatter:on
                        # Skip transforming this instance
                        continue
                    # Extract the single float value of the tensor
                    dequant_softmax = float(scale.item())
                # Insert default scale if the is no dequantizer present
                else:
                    # Default is identity scale
                    dequant_softmax = 1.0

                # The last node of the attention operator is either the detected
                # matmul or the following, optional activation function
                last = act_av_matmul if act_av_matmul is not None else node

                # Tensor names of the threshold inputs
                # Note: order matters
                thresholds = [
                    # TODO: Fix condition once more activation types are
                    #  supported, currently there are only none and thresholds
                    act.input[1] for act in acts if act is not None
                ]

                def act_op_type_str(act):
                    """Convert activation function types to string representation.

                    Only MultiThreshold is supported currently. The attention custom op
                    uses "thresholds" to identify this type.
                    """
                    # Only MultiThreshold is supported currently
                    if act is not None and act.op_type == "MultiThreshold":
                        # The attention custom op uses "thresholds" to identify
                        return "thresholds"
                    # All other types are not supported
                    return "none"

                # The value tensor shape must be compatible with the attention
                # matrix
                assert model.get_tensor_shape(v)[:2] == [qh, kl]

                # Output type of the first matmul
                out_qk_matmul = lhs[-1].output[0]
                # Extend the output type to include the optional thresholding
                # activation
                if act_qk_matmul is not None:
                    # Single output tensor of the activation function
                    out_qk_matmul = act_qk_matmul.output[0]

                def out_bias(act):
                    """Extract output bias of the thresholding activation functions.

                    Only applies to thresholding activations. Extracts via interpreting
                    the node as QONNX custom op.
                    """
                    # Does only apply to thresholding activations
                    if act is not None and act.op_type == "MultiThreshold":
                        # Extract via interpreting the node as QONNX custom op
                        return getCustomOp(act).get_nodeattr("out_bias")
                    # Default bias if no bias
                    return 0.0

                # Fixed node attributes and extracted input/output/initializer
                # tensor names
                kwargs = {
                    # Refer to this operator type by its name
                    "op_type": "ScaledDotProductAttention",
                    # Execution will try to look up the implementation in the
                    # package
                    # referred to by the domain
                    "domain": "finn.custom_op.fpgadataflow",
                    # Execution backend: Required attribute inherited from
                    # HLSCustomOp
                    "backend": "fpgadataflow",
                    # Named inputs and activation thresholds extracted from the
                    # graph pattern
                    "inputs": [q, k, v, *mask, *thresholds],
                    # Named model output extracted from the graph pattern
                    "outputs": last.output,
                    # Set the attribute specifying how to handel the optional
                    # attention mask
                    "mask_mode": mask_mode,
                    # Give node name derived from the operator type and the name
                    # of the triggering node to be removed
                    "name": f"ScaledDotProductAttention_{node.name}"
                }

                # Extract the node attributes of the attention operator from
                # all constituent nodes
                node_attrs = {
                    # Number of attention heads
                    "Heads": qh,
                    # Embedding dimension of queries and keys
                    "QKDim": qe,
                    # Length of the query sequence
                    "QLen": ql,
                    # Embedding dimension of the values
                    "VDim": model.get_tensor_shape(v)[2],
                    # Length of the key and value sequence
                    "KVLen": kl,

                    # Folding along the embedding dimensions
                    # Note: Assume biggest folding possible fitting both
                    # embedding dimensions
                    "EmbFold": math.gcd(qe, model.get_tensor_shape(v)[2]),
                    # Folding along the sequence dimensions
                    # Note: Assume biggest folding possible fitting both
                    # sequence dimensions
                    "SeqFold": math.gcd(ql, kl),

                    # Datatype of query matrix elements
                    "QType": model.get_tensor_datatype(q),
                    # Datatype of key matrix elements
                    "KType": model.get_tensor_datatype(k),
                    # Datatype of value matrix elements
                    "VType": model.get_tensor_datatype(v),
                    # # Datatype of mask matrix elements
                    "MType": mask_dtype.name,
                    # Datatype of attention weights elements
                    "AType": model.get_tensor_datatype(lhs[0].output[0]),
                    # Datatype of output elements
                    "OType": model.get_tensor_datatype(last.output[0]),

                    # Datatype of accumulator elements of the first matmul
                    "AccQKMatMul": model.get_tensor_datatype(lhs[-1].output[0]),
                    # Datatype of output elements of the first matmul
                    # Note: Can be extracted from the left hand side
                    # intermediate outputs
                    "OutQKMatMul": model.get_tensor_datatype(out_qk_matmul),
                    # Activation function type following the first matmul
                    "ActQKMatMul": act_op_type_str(act_qk_matmul),
                    # Output bias to be applied to the thresholding activation
                    # following the Query x Key multiplication
                    "BiasActQKMatMul": out_bias(act_qk_matmul),

                    # Datatype of accumulator elements of the second matmul
                    "AccAVMatMul": model.get_tensor_datatype(node.output[0]),
                    # Datatype of output elements of the second matmul
                    # Note: Always the same as the OType
                    "OutAVMatMul": model.get_tensor_datatype(last.output[0]),
                    # Activation function type following the second matmul
                    "ActAVMatMul": act_op_type_str(act_av_matmul),
                    # Output bias to be applied to the thresholding activation
                    # following the Attention x Value multiplication
                    "BiasActAVMatMul": out_bias(act_av_matmul),

                    # Softmax may be preceded by a de-quantizer scalar
                    # multiplication
                    "DequantSoftmax": dequant_softmax,
                    # Datatype of softmax normalization before applying
                    # activation or type cast. This is called Acc to stick to
                    # the naming scheme of the MatMul operators before.
                    #   Note: Currently this is ALWAYS floats
                    "AccASoftmax": "FLOAT32",
                    # Activation function type following the softmax
                    # normalization of the attention weights
                    "ActASoftmax": act_op_type_str(act_a_softmax),
                    # Output bias to be applied to the thresholding activation
                    # following the softmax normalization of the attention
                    # weights
                    "BiasActASoftmax": out_bias(act_a_softmax),
                }

                def maybe_name(value):
                    """Convert QONNX datatypes to their name (as a string).

                    All QONNX datatypes are instances of BaseDataType. Everything else
                    is assumed to be in the right format already.
                    """
                    # All QONNX datatypes are instances of the BaseDataType
                    if isinstance(value, BaseDataType):
                        # Convert to the name by referring to the datatypes name
                        # attribute
                        return value.name
                    # Everything else is just assumed to be in the right format
                    return value

                # Convert all node attributes DataTypes to string
                # representations of their names
                node_attrs = {
                    key: maybe_name(value) for key, value in node_attrs.items()
                }

                # Create a new custom node replacing the scaled dot-product
                # attention pattern
                attention = oh.make_node(**kwargs, **node_attrs)
                # Insert the new node into the graph
                graph.node.insert(index, attention)
                # Collect all nodes comprising the original pattern
                nodes = [node, transpose, *lhs, act_av_matmul]
                # Remove all nodes of the original pattern
                for n in nodes:
                    # Do not try to remove non-existing nodes
                    if n is not None:
                        graph.node.remove(n)
                # The graph has been modified
                graph_modified = True
        # After rewiring need to re-do the shape annotations
        model = model.transform(InferShapes())  # noqa: Shadows model
        # As attention mask datatype might have been changed, it might be
        # necessary to re-do the datatype annotations
        model = model.transform(InferDataTypes())
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified


class AbsorbMultiThresholdIntoScaledDotProductAttention(Transformation):
    """Absorb a MultiThreshold into ScaledDotProductAttention if there is not
    already an activation included."""

    def apply(self, model: ModelWrapper):  # noqa
        """Apply the transform to a whole model graph."""
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # Any MultiThreshold is a candidate node
            if node.op_type == "MultiThreshold":
                # Cannot be a join-node
                if model.is_join_node(node):
                    # Softly skip transforming this node
                    continue
                # Now we know there is only one producer operation preceding the
                # multi-threshold node
                attention = model.find_direct_predecessors(node)
                # The first node in the graph might have no predecessor
                if attention is None:
                    # Skip this node
                    continue
                # Unpack the single predecessor from the list
                attention = attention[0]
                # Predecessor must actually be a ScaledDotProductAttention for
                # this transform to apply
                if not attention.op_type == "ScaledDotProductAttention":
                    # Skip transforming this instance, probably no need to warn
                    continue
                # The attention operation may not fork for this transformation
                # to be applicable
                if model.is_fork_node(attention):
                    # Softly skip transforming this, will result in standalone
                    # thresholds
                    continue

                # Check whether the attention operation already has an output
                # activation
                if getCustomOp(attention).get_nodeattr("ActAVMatMul") != "none":
                    # Issue a warning to make the user aware of this mismatch
                    # pattern
                    # @formatter:off
                    log.warning(
                        f"{self.__class__.__name__}: Skipping near match: "
                        f" {attention.name} already has an activation:"
                        f" {get_by_name(attention.attribute, 'ActAVMatMul').s}"
                    )
                    # @formatter:on
                    # Skip transforming this instance
                    continue

                # Datatype of the thresholding output, which will be the new
                # output datatype of the attention operator
                dtype = getCustomOp(node).get_nodeattr("out_dtype")
                # Output bias after the thresholding, needs to be absorbed into
                # the attention operator as well
                out_bias = getCustomOp(node).get_nodeattr("out_bias")

                # Collect new attributes
                attrs = {
                    # Datatype of output elements of the second matmul
                    # Note: Always the same as the OType
                    "OutAVMatMul": dtype,
                    # Attention operator output type must be the same as the
                    # output type of the last matmul
                    "OType": dtype,
                    # Activation function type following the second matmul
                    "ActAVMatMul": "thresholds",
                    # Output bias to be applied to the thresholding activation
                    # following the Attention x Value multiplication
                    "BiasActAVMatMul": out_bias,
                }

                # Run over all attributes to be changed
                for key, value in attrs.items():
                    # Remove the existing attribute
                    remove_by_name(attention.attribute, key)
                    # Insert a new attribute with the same name
                    attention.attribute.append(oh.make_attribute(key, value))

                # Append the new threshold tensor as the last input
                attention.input.append(node.input[1])
                # Annotate the new thresholds tensor datatype
                model.set_tensor_datatype(
                    node.input[1], model.get_tensor_datatype(node.input[0])
                )
                # Rewire the output of the attention operator to skip the
                # thresholds node
                attention.output[0] = node.output[0]
                # Remove the thresholding node
                graph.node.remove(node)
                # The graph has been modified
                graph_modified = True
                # Break the loop after adding and removing nodes to start over
                # with a clean index
                break
        # After rewiring need to re-do the shape annotations
        model = model.transform(InferShapes())  # noqa: Shadows model
        # As attention mask datatype might have been changed, it might be
        # necessary to re-do the datatype annotations
        model = model.transform(InferDataTypes())
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified



# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
import os
import shutil
from qonnx.util.basic import roundup_to_integer_multiple

from finn.custom_op.fpgadataflow.fmpadding import FMPadding
from finn.custom_op.fpgadataflow.rtlbackend import RTLBackend


class FMPadding_rtl(FMPadding, RTLBackend):
    """CustomOp wrapper for the finn-rtllib fmpadding_axi component
    Supports adjusting the padding amount and spatial feature sizes at
    runtime."""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {
            # Enable reprogrammable implementation to change FM dimensions,
            # stride, or dilation during runtime
            "dynamic_mode": ("i", False, 0, {0, 1}),
        }
        my_attrs.update(FMPadding.get_nodeattr_types(self))
        my_attrs.update(RTLBackend.get_nodeattr_types(self))
        return my_attrs

    def get_verilog_top_module_intf_names(self):
        # Overload default HLSCustomOp implementation to add axilite control IF
        intf_names = super().get_verilog_top_module_intf_names()
        if self.get_nodeattr("dynamic_mode"):
            intf_names["axilite"] = ["s_axilite"]
        return intf_names

    def get_template_values(self, ifm_dims, pads, chans, simd, idt):
        dimY, dimX = ifm_dims
        padT, padL, padB, padR = pads
        y_counter_bits = int(math.ceil(math.log2(padT + dimY + padB + 1)))
        x_counter_bits = int(math.ceil(math.log2(padL + dimX + padR + 1)))
        topname = self.get_verilog_top_module_name()
        stream_bits = idt.bitwidth() * simd
        stream_bits = int(roundup_to_integer_multiple(stream_bits, 8))
        code_gen_dict = {
            "XCOUNTER_BITS": int(x_counter_bits),
            "YCOUNTER_BITS": int(y_counter_bits),
            "NUM_CHANNELS": int(chans),
            "SIMD": int(simd),
            "ELEM_BITS": idt.bitwidth(),
            "TOP_MODULE_NAME": topname,
            "INIT_XON": int(padL),
            "INIT_XOFF": int(padL + dimX),
            "INIT_XEND": int(padL + dimX + padR - 1),
            "INIT_YON": int(padT),
            "INIT_YOFF": int(padT + dimY),
            "INIT_YEND": int(padT + dimY + padB - 1),
            "STREAM_BITS": int(stream_bits),
        }
        return code_gen_dict

    def get_dynamic_config(self, ifm_dims=None, pads=None):
        """Returns a configuration dict to re-configure FM dimension and
        padding amounts during runtime."""

        if ifm_dims is None:
            ifm_dims = self.get_nodeattr("ImgDim")
        if pads is None:
            pads = self.get_nodeattr("Padding")
        chans = self.get_nodeattr("NumChannels")
        simd = self.get_nodeattr("SIMD")
        idt = self.get_input_datatype()
        code_gen_dict = self.get_template_values(ifm_dims, pads, chans, simd, idt)
        config = {
            "XON": (0 * 4, (code_gen_dict["INIT_XON"])),
            "XOFF": (1 * 4, (code_gen_dict["INIT_XOFF"])),
            "XEND": (2 * 4, (code_gen_dict["INIT_XEND"])),
            "YON": (3 * 4, (code_gen_dict["INIT_YON"])),
            "YOFF": (4 * 4, (code_gen_dict["INIT_YOFF"])),
            "YEND": (5 * 4, (code_gen_dict["INIT_YEND"])),
        }
        return config

    def generate_hdl(self, model, fpgapart, clk):
        rtlsrc = os.path.join(os.environ["FINN_RTLLIB"], "fmpadding/hdl")
        template_path = rtlsrc + "/fmpadding_template.v"
        dims = self.get_nodeattr("ImgDim")
        pads = self.get_nodeattr("Padding")
        chans = self.get_nodeattr("NumChannels")
        simd = self.get_nodeattr("SIMD")
        idt = self.get_input_datatype()
        code_gen_dict = self.get_template_values(dims, pads, chans, simd, idt)
        # save top module name so we can refer to it after this node has been renamed
        # (e.g. by GiveUniqueNodeNames(prefix) during MakeZynqProject)
        self.set_nodeattr("gen_top_module", self.get_verilog_top_module_name())

        # apply code generation to templates
        code_gen_dir = self.get_nodeattr("code_gen_dir_ipgen")
        with open(template_path, "r") as f:
            template = f.read()
        for key_name in code_gen_dict:
            key = "$%s$" % key_name
            template = template.replace(key, str(code_gen_dict[key_name]))

        with open(
            os.path.join(code_gen_dir, self.get_verilog_top_module_name() + ".v"),
            "w",
        ) as f:
            f.write(template)

        sv_files = ["fmpadding_axi.sv", "fmpadding.sv", "axi2we.sv"]
        for sv_file in sv_files:
            shutil.copy(rtlsrc + "/" + sv_file, code_gen_dir)
        # set ipgen_path and ip_path so that HLS-Synth transformation
        # and stich_ip transformation do not complain
        self.set_nodeattr("ipgen_path", code_gen_dir)
        self.set_nodeattr("ip_path", code_gen_dir)

    def get_rtl_file_list(self, abspath=False):
        if abspath:
            code_gen_dir = self.get_nodeattr("code_gen_dir_ipgen") + "/"
            rtllib_dir = os.path.join(os.environ["FINN_RTLLIB"], "fmpadding/hdl/")
        else:
            code_gen_dir = ""
            rtllib_dir = ""

        verilog_files = [
            rtllib_dir + "fmpadding_axi.sv",
            rtllib_dir + "fmpadding.sv",
            rtllib_dir + "axi2we.sv",
            code_gen_dir + self.get_nodeattr("gen_top_module") + ".v",
        ]
        return verilog_files

    def code_generation_ipi(self):
        """Constructs and returns the TCL for node instantiation in Vivado IPI."""
        code_gen_dir = self.get_nodeattr("code_gen_dir_ipgen")

        sourcefiles = [
            "fmpadding_axi.sv",
            "fmpadding.sv",
            "axi2we.sv",
            self.get_nodeattr("gen_top_module") + ".v",
        ]

        sourcefiles = [os.path.join(code_gen_dir, f) for f in sourcefiles]

        cmd = []
        for f in sourcefiles:
            cmd += ["add_files -norecurse %s" % (f)]
        cmd += [
            "create_bd_cell -type module -reference %s %s"
            % (self.get_nodeattr("gen_top_module"), self.onnx_node.name)
        ]
        return cmd

    def execute_node(self, context, graph):
        mode = self.get_nodeattr("exec_mode")
        if mode == "cppsim":
            FMPadding.execute_node(self, context, graph)
        elif mode == "rtlsim":
            RTLBackend.execute_node(self, context, graph)



"""Transformer benchmarking module adapted from Christoph's attention-dummy repository."""

# PyTorch base package: Math and Tensor Stuff
import json
import numpy as np
import random
import torch
from brevitas.export import export_qonnx

# Brevitas: Quantized versions of PyTorch layers
from brevitas.nn import (
    QuantEltwiseAdd,
    QuantIdentity,
    QuantLinear,
    QuantMultiheadAttention,
    QuantReLU,
)

# Brevitas wrapper around PyTorch tensors adding quantization information
from brevitas.quant_tensor import QuantTensor
from qonnx.core.modelwrapper import ModelWrapper

# Range information structure for seeding the range analysis for converting
# quantized activations to MultiThreshold
from qonnx.util.range_analysis import RangeInfo

# Progressbar
from tqdm import trange

# FINN dataflow builder
import finn.builder.build_dataflow_config as build_cfg
from finn.benchmarking.bench_base import bench

# Custom build steps required to streamline and convert the attention operator
from finn.builder.custom_step_library.transformer import (
    prepare_graph,
    set_fifo_depths,
    set_target_parallelization,
    step_apply_folding_config,
    step_convert_attention_to_hw,
    step_convert_depth_wise_to_hw,
    step_convert_elementwise_binary_to_hw,
    step_convert_lookup_to_hw,
    step_convert_split_concat_to_hw,
    step_replicate_streams,
    step_streamline,
)


# ADAPTED FROM utils.py
def seed(s):
    """Seeds all relevant random number generators to the same seed for reproducibility."""
    random.seed(s)
    np.random.seed(s)
    torch.manual_seed(s)


# ADAPTED FROM model.py
def weight_quantizer(bits, _signed=True):
    """Derives a weight quantizer, leaving bit-width and signedness configurable."""
    # Brevitas quantizer base classes
    from brevitas.inject.enum import RestrictValueType
    from brevitas.quant.base import MaxStatsScaling, NarrowIntQuant
    from brevitas.quant.solver import WeightQuantSolver

    # Derive a Quantizer from the brevitas bases
    class Quantizer(NarrowIntQuant, MaxStatsScaling, WeightQuantSolver):
        """Weight quantizer with configurable bit-width and signedness."""

        # Configure the quantization bit-width
        bit_width = bits
        # Signedness of the quantization output
        signed = _signed
        # Per tensor quantization, not per channel
        scaling_per_output_channel = False
        # What is this? Copied from PerTensorFloatScaling*
        #   Probably restricts the scale to be floating-point?
        restrict_scaling_type = RestrictValueType.FP

    # Return the derived quantizer configuration
    return Quantizer


def bias_quantizer(bits, _signed=True):
    """Derives a bias quantizer, leaving bit-width and signedness configurable."""
    # Brevitas quantizer base classes
    from brevitas.quant import IntBias

    # Derive a Quantizer from the brevitas bases
    class Quantizer(IntBias):
        """Bias quantizer with configurable bit-width and signedness."""

        # Configure the quantization bit-width
        bit_width = bits
        # Signedness of the quantization output
        signed = _signed
        # Do not require the bit-width to be adjusted to fit the accumulator to
        # which the bias is added
        requires_input_bit_width = False

    # Return the derived quantizer configuration
    return Quantizer


def act_quantizer(bits, _signed=True):
    """Derives an activation quantizer, leaving bit-width and signedness configurable."""
    # Brevitas quantizer base classes
    from brevitas.inject.enum import RestrictValueType
    from brevitas.quant.base import IntQuant, ParamFromRuntimePercentileScaling
    from brevitas.quant.solver import ActQuantSolver

    # Derive a Quantizer from the brevitas bases
    class Quantizer(IntQuant, ParamFromRuntimePercentileScaling, ActQuantSolver):
        """Activation quantizer with configurable bit-width and signedness."""

        # Configure the quantization bit-width
        bit_width = bits
        # Signedness of the quantization output
        signed = _signed
        # Per tensor quantization, not per channel
        scaling_per_output_channel = False
        # What is this? Copied from PerTensorFloatScaling*
        #   Probably restricts the scale to be floating-point?
        restrict_scaling_type = RestrictValueType.FP

    # Return the derived quantizer configuration
    return Quantizer


def get_norm(key, normalized_shape):
    """Gets the normalization layer from configuration key."""

    class Transpose(torch.nn.Module):
        """Transposes Sequence and Embedding dimensions."""

        def forward(self, x):  # noqa: May be static
            """Forward pass transposing the feature map."""
            # Transpose the last two dimensions of batch x seq x emb layout
            return torch.transpose(x, dim0=-1, dim1=-2)

    # Dictionary mapping keys to supported normalization layer implementations
    norms = {
        # PyTorch default layer normalization. Needs to know the shape of the
        # feature map to be normalized
        "layer-norm": torch.nn.LayerNorm(
            # Note: Disable affine parameters as potential negative scale causes
            # streamlining issues later
            normalized_shape=normalized_shape,
            elementwise_affine=False,
        ),
        # PyTorch default 1-dimensional batch normalization. Needs to transpose
        # embedding and sequence dimension to normalized over the embedding
        # dimension, which is expected to be second.
        "batch-norm": torch.nn.Sequential(
            # Note: Disable affine parameters as potential negative scale causes
            # streamlining issues later
            Transpose(),
            torch.nn.LazyBatchNorm1d(affine=False),
            Transpose(),
        ),
        # No normalization by a PyTorch built-in identity layer. Should not
        # appear in the graph.
        "none": torch.nn.Identity(),
    }

    # Select the normalization layer by key
    return norms[key]


def get_mask(key, length):
    """Gets the attention mask from configuration key and shape."""
    # Dictionary mapping keys to supported normalization layer implementations
    masks = {
        # No attention mask
        "none": None,
        # Generate the upper triangular mask for causal attention
        "causal": torch.nn.Transformer.generate_square_subsequent_mask(length),
        # Square matrix with entries randomly set to -inf or 0.0 with 50%
        # probability each
        "random": torch.where(  # noqa: Confused by types?
            torch.rand(length, length) > 0.5, -torch.inf, 0.0
        ),
    }
    # Select the mask type by key
    return masks[key]


class TransformerBlock(torch.nn.Module):
    """Single-layer scaled dot-product attention block with MLP and normalization."""

    def __init__(self, num_heads, emb_dim, mlp_dim, seq_len, bias, norm, mask, bits):
        """Initializes the model and registers the module parameters."""
        # Initialize the PyTorch Module superclass
        super().__init__()

        # Input quantizer to the scaled dot-product attention operations, shared
        # by queries, keys and values inputs. It is important to have this
        # quantizer separate and not preceding the fork node of the residual
        # branches to avoid consecutive quantizers in the skip branch.
        # Note: For some reason it seems not to be possible to use the
        #   in_proj_input_quant of the attention operator
        self.sdp_input_quant = QuantIdentity(
            # Quantize at the output
            act_quant=act_quantizer(bits, _signed=True),
            # Pass quantization information on to the next layer.
            return_quant_tensor=True,
        )
        # Quantized scaled dot-product attention operator
        self.sdp = QuantMultiheadAttention(
            # Size of the embedding dimension (input and output)
            embed_dim=emb_dim,
            # Number of attention heads
            num_heads=num_heads,
            # Enable a bias added to the input and output projections
            bias=bias,
            # Layout of the inputs:
            #   Batch x Sequence x Embedding (batch-first, True)
            #   Sequence x Batch x Embedding (batch-second, False)
            batch_first=True,
            # If query, key and value input are the same, packed input
            # projections use a single, large linear projection to produce
            # the actual query, key and value inputs. Otherwise, use
            # separate linear projections on each individual input.
            packed_in_proj=False,
            # Brevitas has this as an unsigned quantizer by default, but
            # finn can only handle signed quantizer
            attn_output_weights_quant=act_quantizer(bits, _signed=True),
            # Insert an additional quantizer in front ot the softmax. In our
            # finn custom-op, this will be matched to the quantizer
            # following the query and key matmul.
            # Note: Disable to prevent the quantizer from tripping over -inf
            # from the attention mask
            softmax_input_quant=None,
            # Quantize the input projections weights as configured
            in_proj_weight_quant=weight_quantizer(bits, _signed=True),
            # Quantize the bias of the input projections as configured
            in_proj_bias_quant=bias_quantizer(bits, _signed=True),
            # No quantization in front of the input projections as this is
            # either done by a standalone quantizer preceding the whole block
            in_proj_input_quant=None,
            # Quantize the output projections weights as configured
            out_proj_weight_quant=weight_quantizer(bits, _signed=True),
            # Quantize the bias of the output projections as configured
            out_proj_bias_quant=bias_quantizer(bits, _signed=True),
            # Quantize the input to the output projection as configured
            out_proj_input_quant=act_quantizer(bits, _signed=True),
            # Quantizer the key after projections as configured
            k_transposed_quant=act_quantizer(bits, _signed=True),
            # Quantize the queries after projections as configured
            q_scaled_quant=act_quantizer(bits, _signed=True),
            # Quantize the values after projection as configured
            v_quant=act_quantizer(bits, _signed=True),
            # No output quantization for now, as stacking multiple layers
            # results in multiple multi-thresholds in succession
            out_proj_output_quant=None,
            # Return the quantization parameters so the next layer can
            # quantize the bias
            return_quant_tensor=True,
        )
        # Residual branch addition skipping over the attention layer
        self.residual_sdp = QuantEltwiseAdd(
            # Shared input activation quantizer such that the scales at both
            # input branches are identical. This allows floating point scale
            # factor to be streamlined past the add-node.
            input_quant=act_quantizer(bits, _signed=True),
            # Disable the output quantizer after the add operation. Output of
            # the add will have one more bit than the inputs, which is probably
            # fine and does not require re-quantization.
            output_quant=None,
            # Pass quantization information on to the next layer.
            return_quant_tensor=True,
        )
        # Normalization following the attention layer
        self.norm_sdp = torch.nn.Sequential(
            # Select the normalization layer implementation
            get_norm(key=norm, normalized_shape=emb_dim),
            # No quantizer to avoid consecutive quantizer in the MLP residual
            # branch. See input quantizer in front of the first MLP layer.
        )

        # Quantized MLP following the scaled dot-product attention
        self.mlp = torch.nn.Sequential(
            # Quantize the inputs to the MLP block. Placed here to not have this
            # at the input of the residual branch.
            QuantIdentity(
                # Quantize at the output
                act_quant=act_quantizer(bits, _signed=True),
                # Pass quantization information on to the next layer.
                return_quant_tensor=True,
            ),
            # First mlp layer projecting to the mlp dimension
            QuantLinear(
                # Inputs have the size of the attention embedding dimension
                emb_dim,
                # Project to the configured mlp dimension, which is typically
                # larger than the embedding dimension
                mlp_dim,
                # Enable the learned bias vector
                bias=bias,
                # Quantize weights to the same representation as all other
                # layers
                weight_quant=weight_quantizer(bits, _signed=True),
                # Quantize the bias to the same representation as all other
                # layers
                bias_quant=bias_quantizer(bits, _signed=True),
                # No input quantizer as this is directly preceded by a
                # standalone quantizer
                input_quant=None,
                # Not output quantizer as this is directly followed by a
                # quantized ReLU activation taking care of quantization
                output_quant=None,
                # Return the quantization parameters so the next layer can
                # quantize the bias
                return_quant_tensor=True,
            ),
            # Use the ReLU activation function instead of the more commonly used
            # GELU, as the latter is not mapped easily to hardware with FINN
            QuantReLU(
                # Note: ReLU must be quantized to unsigned representation
                act_quant=act_quantizer(bits, _signed=False),
                # Return the quantization parameters so the next layer can
                # quantize the bias
                return_quant_tensor=True,
            ),
            # Second mlp layer projecting back to the embedding dimension
            QuantLinear(
                # Inputs have the configured mlp dimension, which is typically
                # larger than the embedding dimension
                mlp_dim,
                # Project back to the size of the attention embedding dimension
                emb_dim,
                # Enable the learned bias vector
                bias=bias,
                # Quantize weights to the same representation as all other
                # layers
                weight_quant=weight_quantizer(bits, _signed=True),
                # Quantize the bias to the same representation as all other
                # layers
                bias_quant=bias_quantizer(bits, _signed=True),
                # No input quantizer as the inputs are already quantized by the
                # preceding ReLU layer
                input_quant=None,
                # Not output quantizer as this is directly followed by a
                # quantized element-wise addition taking care of quantization
                output_quant=None,
                # Pass quantization information on to the next layer.
                return_quant_tensor=True,
            ),
        )
        # Residual branch addition skipping over the MLP layer
        self.residual_mlp = QuantEltwiseAdd(
            # Shared input activation quantizer such that the scales at both
            # input branches are identical. This allows floating point scale
            # factor to be streamlined past the add-node.
            input_quant=act_quantizer(bits, _signed=True),
            # Disable the output quantizer after the add operation. Output of
            # the add will have one more bit than the inputs, which is probably
            # fine and does not require re-quantization.
            output_quant=None,
            # Pass quantization information on to the next layer.
            # Note: Not for the last layer to allow this to be combined with
            # standard pytorch calls like .detach() or .numpy(), which are
            # not directly available on QuantTensor.
            return_quant_tensor=True,
        )
        # Normalization following the attention layer
        self.norm_mlp = torch.nn.Sequential(
            # Select the normalization layer implementation
            get_norm(key=norm, normalized_shape=emb_dim),
            # No quantizer to avoid consecutive quantizer in the SDP residual
            # branch
        )
        # Generate the attention mask according to configuration
        self.mask = get_mask(mask, seq_len)

    # Forward pass adding positional encoding to the input tensor
    def forward(self, x):
        """Forward pass adding sinusoidal positional encoding to input tensor."""
        """Forward pass through the transformer block."""
        # Move the mask to the same device as the input, just in case...
        mask = self.mask.to(x.device) if self.mask is not None else None
        # Quantize the input to the attention block
        q = self.sdp_input_quant(x)
        # Scaled dot-product attention with residual branch and normalization
        x = self.norm_sdp(self.residual_sdp(x, self.sdp(q, q, q, attn_mask=mask)[0]))
        # MLP layer with residual branch and normalization
        return self.norm_mlp(self.residual_mlp(x, self.mlp(x)))


class QuantSinusoidalPositionalEncoding(torch.nn.Module):
    """Quantized sinusoidal positional encoding layer."""

    def __init__(self, input_quant, output_quant, return_quant_tensor):
        """Initializes the model and registers the module parameters."""
        # Initialize the PyTorch Module superclass
        super().__init__()
        # Adds the quantized input and positional encoding
        self.add = QuantEltwiseAdd(
            # Input quantization to be applied to the input as well as the
            # positional encodings
            input_quant=input_quant,
            # Quantize the outputs after adding input and positional encoding
            output_quant=output_quant,
            # Returns quantization information to the next layer
            return_quant_tensor=return_quant_tensor,
        )

    def forward(self, x):
        """Forward pass adding sinusoidal positional encoding to input tensor."""
        # Get the size of the inputs to dynamically generate encodings of the
        # same size
        _, seq, emb = x.shape
        # Start by enumerating all steps of the sequence
        i = torch.as_tensor([[n] for n in range(seq)])
        # Scale factor adjusting the frequency/wavelength of the sinusoid
        # depending on the embedding dimension index
        f = torch.as_tensor([1e4 ** -(i / emb) for i in range(0, emb, 2)])
        # Prepare empty positional encoding tensor of the same size as the input
        pos = torch.empty(seq, emb)
        # Fill the positional encoding with alternating sine and cosine waves
        pos[:, 0::2] = torch.sin(f * i)
        pos[:, 1::2] = torch.cos(f * i)
        # Move the encoding tensor to the same device as the input tensor
        pos = pos.to(x.device, dtype=x.dtype)
        # Add the quantized encoding to the quantized input
        return self.add(x, pos)


class QuantLearnedPositionalEncoding(torch.nn.Module):
    """Quantized learned positional encoding layer."""

    def __init__(self, seq_len, emb_dim, input_quant, output_quant, return_quant_tensor):
        """Initializes the model and registers the module parameters."""
        # Initialize the PyTorch Module superclass
        super().__init__()
        # Adds the quantized input and positional encoding
        self.add = QuantEltwiseAdd(
            # Input quantization to be applied to the input as well as the
            # positional encodings
            input_quant=input_quant,
            # Quantize the outputs after adding input and positional encoding
            output_quant=output_quant,
            # Returns quantization information to the next layer
            return_quant_tensor=return_quant_tensor,
        )
        # Register a parameter tensor representing the not quantized positional
        # encoding
        self.pos = torch.nn.Parameter(torch.empty(seq_len, emb_dim))
        # Reset/Initialize the parameter tensor
        self.reset_parameters()

    # Resets/Initializes the positional encoding parameter tensor
    def reset_parameters(self):
        """Resets/Initializes positional encoding parameters from normal distribution."""
        # Initialize the positional encoding from a normal distribution with
        # zero mean and unit standard deviation
        torch.nn.init.normal_(self.pos, mean=0, std=1)

    # Forward pass adding positional encoding to the input tensor
    def forward(self, x):
        """Forward pass adding learned positional encoding to input tensor."""
        # Add the quantized encoding to the quantized input
        return self.add(x, self.pos)


class LazyQuantLearnedPositionalEncoding(
    torch.nn.modules.lazy.LazyModuleMixin, QuantLearnedPositionalEncoding  # noqa
):
    """
    Lazy version of the learned encoding not requiring input dimensions at initialization,
    inferring these at the first forward pass.
    """

    # Once initialized, this will become a QuantLearnedPositionalEncoding as
    # defined above
    cls_to_become = QuantLearnedPositionalEncoding
    # Parameter tensor of the QuantLearnedPositionalEncoding is uninitialized
    pos: torch.nn.UninitializedParameter

    # Initializes the model and registers the module parameters
    def __init__(self, input_quant, output_quant, return_quant_tensor):
        """Initializes lazy learned positional encoding with uninitialized parameters."""
        # Initialize the quantizer parts of QuantLearnedPositionalEncoding,
        # leaving the dimensions empty
        super().__init__(0, 0, input_quant, output_quant, return_quant_tensor)
        # Register an uninitialized parameter tensor for the positional encoding
        self.pos = torch.nn.UninitializedParameter()

    # Resets/Initializes the positional encoding parameter tensor
    def reset_parameters(self):
        """Resets parameters if already initialized, otherwise delegates to parent."""
        # If this has already been initialized, delegate to the actual
        # implementation
        if not self.has_uninitialized_params():
            super().reset_parameters()

    # Initializes/Materializes the uninitialized parameter tensor given some
    # sample input tensor to infer the dimensions
    def initialize_parameters(self, x):
        """Materializes uninitialized parameter tensor by inferring dimensions from input."""
        # Only materialize the parameter tensor if it is not yet initialized
        if self.has_uninitialized_params():
            # Do not accumulate gradient information from initialization
            with torch.no_grad():
                # Get the size of the inputs to generate encodings of the same
                # size
                _, seq, emb = x.shape
                # Materialize the positional encoding parameter tensor
                self.pos.materialize((seq, emb))
                # Properly initialize the parameters by resetting the values
                self.reset_parameters()


class QuantBinaryPositionalEncoding(torch.nn.Module):
    """Quantized binary positional encoding layer."""

    # Initializes the model and registers the module parameters
    def __init__(self, input_quant, output_quant, return_quant_tensor):
        """Initializes binary positional encoding with quantized addition."""
        # Initialize the PyTorch Module superclass
        super().__init__()
        # Adds the quantized input and positional encoding
        self.add = QuantEltwiseAdd(
            # Input quantization to be applied to the input as well as the
            # positional encodings
            input_quant=input_quant,
            # Quantize the outputs after adding input and positional encoding
            output_quant=output_quant,
            # Returns quantization information to the next layer
            return_quant_tensor=return_quant_tensor,
        )

    # Forward pass adding positional encoding to the input tensor
    def forward(self, x):
        """Forward pass adding binary positional encoding to input tensor."""
        # Get the size of the inputs to dynamically generate encodings of the
        # same size
        _, seq, emb = x.shape
        # Binary positional encoding fills the embedding dimension with the bit
        # pattern corresponding to the position in the sequence
        pos = torch.as_tensor([[(n & (1 << bit)) >> bit for bit in range(emb)] for n in range(seq)])
        # Move the encoding tensor to the same device as the input tensor
        pos = pos.to(x.device, dtype=x.dtype)
        # Add the quantized encoding tp the quantized input
        #   Note: Convert encoding to bipolar representation
        return self.add(x, 2 * pos - 1)


def get_positional_encoding(key, input_quant, output_quant, return_quant_tensor):
    """Gets the positional encoding layer from configuration key, quantizers and shape."""
    # Dictionary mapping keys to supported normalization layer implementations
    masks = {
        # No positional encoding
        "none": QuantIdentity(act_quant=input_quant, return_quant_tensor=return_quant_tensor),
        # Fixed, sinusoidal positional encoding according to Vaswani et al. with
        # added quantizers
        "sinusoidal": QuantSinusoidalPositionalEncoding(
            input_quant, output_quant, return_quant_tensor
        ),
        # Fixed, binary positional encoding with quantizers
        "binary": QuantBinaryPositionalEncoding(input_quant, output_quant, return_quant_tensor),
        # Learned positional encoding with quantizers
        "learned": LazyQuantLearnedPositionalEncoding(
            input_quant, output_quant, return_quant_tensor
        ),
    }
    # Select the positional encoding type by key
    return masks[key]


def unpack_from_quant(tensor: torch.Tensor | QuantTensor):
    """Unpacks the standard PyTorch tensor from a brevitas QuantTensor."""
    # If this is a QuantTensor we can extract the wrapped tensor
    if isinstance(tensor, QuantTensor):
        # The underlying tensor is wrapped as the value attribute
        return tensor.value
    # Assume this is already a plain PyTorch tensor
    return tensor


class DummyTransformer(torch.nn.Module):
    """Dummy transformer encoder model."""

    def __init__(
        self,
        # Number of layers of attention blocks
        num_layers,
        # Number of attention heads per block
        num_heads,
        # Size of embedding dimension going into/out of the attention block
        emb_dim,
        # Size of MLP dimension in each attention block
        mlp_dim,
        # Length of the input sequence, i.e., context size
        seq_len,
        # Enables bias term added to Linear layers
        bias,
        # Quantization bit-width: For now all layers are quantized to the
        # same bit-width
        bits,
        # Type of normalization layer to use in the transformer blocks
        #   Options are: layer-norm, batch-norm and none
        norm="none",
        # Type of attention mask to use
        #   Options are: none, causal or const
        mask="none",
        # Type of positional encoding to use at the input
        #   Options are: none, sinusoidal, binary, learned
        positional_encoding="none",
    ):
        """Initializes the model and registers the module parameters."""
        # Initialize the PyTorch Module superclass
        super().__init__()

        # Positional encoding layer at the input
        self.pos = get_positional_encoding(
            # Select the implementation by configuration key
            key=positional_encoding,
            # Quantize the inputs to the positional encoding to the same
            # bit-width as the input
            input_quant=act_quantizer(bits, _signed=True),
            # Quantize the sum of input and positional encoding to the same
            # bit-width as the input
            output_quant=None,
            # Pass quantization information on to the next layer
            return_quant_tensor=True,
        )

        # Sequence of num_layers transformer encoder blocks
        self.encoder = torch.nn.Sequential(
            *[
                TransformerBlock(num_heads, emb_dim, mlp_dim, seq_len, bias, norm, mask, bits)
                for _ in range(num_layers)
            ]
        )

    def forward(self, x):
        """Model forward pass."""
        # Add positional encoding to the input and feed through the encoder
        # stack
        # Note: Get the wrapped value out of the QuantTensor to have only a
        # single output from the model.
        return unpack_from_quant(self.encoder(self.pos(x)))


# ADAPTED FROM export.py


def is_norm_layer(module):
    """Check whether a layer is a normalization layer of some supported type."""
    # Set of normalization layer (bases) which maybe need to be patched
    norm_layers = {
        # All BatchNorm and InstanceNorm variants derive from this baseclass
        torch.nn.modules.batchnorm._NormBase,  # noqa: Access to _NormBase
        # LayerNorm has a unique implementation
        torch.nn.LayerNorm,
    }
    # Check the module against all supported norm layer types
    return any(isinstance(module, norm) for norm in norm_layers)


def patch_non_affine_norms(model: torch.nn.Module):  # noqa: Shadows model
    """Fixes export issues of normalization layers with disabled affine parameters."""
    # Somehow the export to ONNX trips when it encounters the weight and bias tensor
    # to be 'None'.
    # Iterate all modules in the model container
    for name, module in model.named_modules():
        # If the module is a normalization layer it might require patching the
        # affine parameters
        if is_norm_layer(module):
            # Check whether affine scale parameters are missing
            if hasattr(module, "weight") and module.weight is None:
                # There need to be running statistics to patch the scales
                if hasattr(module, "running_var"):
                    # Patch the affine bias by all 1 tensor of the same shape,
                    # type and device as the running variance
                    module.weight = torch.nn.Parameter(torch.ones_like(module.running_var))
            # Check whether affine bias parameters are missing
            if hasattr(module, "bias") and module.bias is None:
                # There need to be running statistics to patch the scales
                if hasattr(module, "running_mean"):
                    # Patch the affine bias by all 0 tensor of the same shape,
                    # type and device as the running mean
                    module.bias = torch.nn.Parameter(torch.zeros_like(module.running_var))
    # Return the patched model container
    return model


template_folding_yaml = """
# Per operator type default configurations
defaults:
    # Scaled dot-product attention head implemented via HLS
    ScaledDotProductAttention_hls:
        # Type of memory to be used for internal buffer storage
        #   Options: auto, block, distributed, ultra
        ram_style: block
        # Type of memory to be used for threshold storage
        #   Options: auto, block, distributed
        ram_style_thresholds: block
        # Type of memory to be used fo the attention mask (if present)
        #   Options: auto, block, distributed
        ram_style_mask: block
        # Resource type to be used for implementing multiplications/MACs
        #   Options: auto, lut or dsp
        mac_resource: lut
    # Addition of two inputs (constants or streamed) implemented via HLS
    ElementwiseAdd_hls:
        # Type of memory to be used for internal buffer storage and/or constant
        # parameter tensors
        #   Options: auto, block, distributed, ultra
        ram_style: distributed
    # Matrix vector activation unit implemented via HLS
    MVAU_hls:
        # Resource type to be used for implementing multiplications/MACs
        #   Options: auto, lut or dsp
        resType: dsp
        # Memory mode for weight storage
        #   Options: internal_embedded, internal_decoupled, external
        mem_mode: internal_decoupled
        # Type of memory to be used for weight storage if "internal_decoupled"
        #   Options: auto, block, distributed, ultra
        ram_style: block
        # Type of memory to be used for threshold storage
        #   Options: auto, block, distributed
        ram_style_thresholds: block
        # Makes weights writeable through AXI-lite interface at runtime
        runtime_writeable_weights: 0
    # Matrix vector activation unit implemented via RTL
    MVAU_rtl:
        # Resource type to be used for implementing multiplications/MACs
        #   Options: auto, lut or dsp
        # Note: RTL MVAU currently does not support LUT-based implementation
        resType: dsp
        # Memory mode for weight storage
        #   Options: internal_embedded, internal_decoupled, external
        mem_mode: internal_decoupled
        # Type of memory to be used for weight storage if "internal_decoupled"
        #   Options: auto, block, distributed, ultra
        ram_style: block
        # Makes weights writeable through AXI-lite interface at runtime
        runtime_writeable_weights: 0
    # Multi-thresholds implemented via HLS (applies to standalone thresholds)
    Thresholding_hls:
        # Memory mode for threshold storage
        #   Options: internal_embedded, internal_decoupled
        mem_mode: internal_decoupled
        # Type of memory to be used for threshold storage if "internal_decoupled"
        #   Options: distributed, block
        ram_style: distributed
        # Makes thresholds writeable through AXI-lite interface at runtime
        runtime_writeable_weights: 0
    # Multi-thresholds implemented via RTL (applies to standalone thresholds)
    Thresholding_rtl:
        # Decides to use BRAM, URAM or LUTs for threshold memory, depending on the
        # depth of the thresholds
        # Note: This combination forces "distributed" LUT implementation
        depth_trigger_uram: 2147483647  # "infinity"
        depth_trigger_bram: 2147483647  # "infinity"
    #    # Note: This combination forces "block" RAM implementation
    #    depth_trigger_uram: 0
    #    depth_trigger_bram: 1
    #    # Note: This combination forces "ultra" RAM implementation
    #    depth_trigger_uram: 1
    #    depth_trigger_bram: 0
    #    # Note: This combination is equivalent to "auto"
    #    depth_trigger_uram: 0
    #    depth_trigger_bram: 0
        # Makes thresholds writeable through AXI-lite interface at runtime
        runtime_writeable_weights: 0
    # FIFO implemented via RTL (there is no HLS FIFO implementation in FINN)
    StreamingFIFO_rtl:
        # RTL vs. IPI implementation of FIFOs
        #   Options: rtl, vivado
        impl_style: rtl
        # Resource type for FIFOs when impl_style is vivado
        #   Options: auto, block, distributed, ultra
        ram_style: distributed
    # Individual, named node-specific configurations here
    # ...
"""


class bench_transformer(bench):
    """Benchmark class for quantized transformer models."""

    def _step_export_onnx(self, output_onnx_path):
        """Generates a dummy transformer block, not used for actual models (RadioML, GPT, etc.)."""
        # Generates a dummy transformer block,
        # not used for actual models (RadioML, GPT, etc.)

        # Load the parameters file
        # params = dvc.api.params_show("params.yaml")
        # Seed all RNGs
        seed(self._params["seed"])
        # Make PyTorch behave deterministically if possible
        torch.use_deterministic_algorithms(mode=True, warn_only=True)
        # Create a model instance from the configuration parameters
        # model = DummyTransformer(**params["model"])
        model = DummyTransformer(
            num_layers=self._params["model_num_layers"],
            num_heads=self._params["model_num_heads"],
            emb_dim=self._params["model_emb_dim"],
            mlp_dim=self._params["model_mlp_dim"],
            seq_len=self._params["model_seq_len"],
            bias=self._params["model_bias"],
            bits=self._params["model_bits"],
            norm=self._params["model_norm"],
            mask=self._params["model_mask"],
            positional_encoding=self._params["model_positional_encoding"],
        )

        # Get the configured sequence length and embedding dimension to generate
        # test inputs
        seq, dim = self._params["model_seq_len"], self._params["model_emb_dim"]
        # No gradient accumulation for calibration passes required
        with torch.no_grad():
            # Check whether GPU training is available and select the appropriate
            # device
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            # Move the model to the training device
            model = model.to(device)
            # Multiple passes of calibration might be necessary for larger/deep
            # models
            for _ in trange(0, self._params["calibration_passes"], desc="calibrating"):
                # Pass random data through the model to "calibrate" dummy quantizer.
                # Large batch to have more calibration samples. Otherwise, there is
                # too much deviation between this calibration and the verification
                # samples.
                model(torch.rand(128, seq, dim, device=device))
            # Move the model back to the CPU
            model = model.cpu()
        # Prevent export issue for missing affine normalization parameters
        model = patch_non_affine_norms(model)
        # Switch model to evaluation mode to have it fixed for export
        model = model.eval()
        # Sample random input tensor in batch-first layout
        x = torch.rand(1, seq, dim)
        # Compute attention output
        o = model(x)
        # Save the input and output data for verification purposes later
        np.save("inp.npy", x.detach().numpy())
        np.save("out.npy", o.detach().numpy())
        self._build_inputs["input_npy_path"] = "inp.npy"
        self._build_inputs["output_npy_path"] = "out.npy"
        # Export the model graph to QONNX
        # export_qonnx(model, (x,), "attention.onnx", **self._params["export"])
        export_qonnx(model, (x,), output_onnx_path, opset_version=14, do_constant_folding=True)

    def _step_build_setup(self):
        """Set up the dataflow build for transformer models with custom attention steps."""
        # with open("params.yaml") as file:
        #    params = yaml.safe_load(file)
        # Seed all RNGs
        seed(self._params["seed"])
        # Extract sequence length and embedding dimension from parameters
        if "model_seq_len" in self._params and "model_emb_dim" in self._params:
            # for dummy Transformer DUT
            seq_len, emb_dim = self._params["model_seq_len"], self._params["model_emb_dim"]
        else:
            # for real input models
            inp_shape = np.load(self._build_inputs["input_npy_path"]).shape
            if len(inp_shape) == 3:
                # for RadioML Transformers
                _, seq_len, emb_dim = inp_shape
            else:
                # for GPTs (why is this different?)
                model = ModelWrapper(self._build_inputs["onnx_path"])
                _, seq_len, emb_dim = model.get_tensor_shape(
                    "/emb_add/input_quant/export_handler/Quant_output_0"
                )

        # Read the input value range information for the dataset from the parameters
        # Note: Consider calibrating this on the fly from the dataset
        value_range = [-100, +100]  # params["build"]["range"] # TODO: make configurable?
        input_range = tuple(np.array([value_range]).T)
        # Construct the seed range information of the input tensor
        range_info = RangeInfo(shape=(1, seq_len, emb_dim), range=input_range)

        # Prepare config files
        # TODO: make configurable
        # TODO: log intermediate files such as inp.npy, folding.yaml,
        # or specialize_layers.jon as artifacts, maybe create in unique temp dirs
        specialize_layers_dict = {
            "Defaults": {"preferred_impl_style": ["rtl", ["MVAU", "Thresholding"]]},
            "": {"preferred_impl_style": ""},
        }
        with open("specialize_layers.json", "w") as f:
            json.dump(specialize_layers_dict, f, indent=2)
        with open("folding.yaml", "w") as f:
            f.write(template_folding_yaml)

        # Create a configuration for building the scaled dot-product attention
        # operator to a hardware accelerator
        cfg = build_cfg.DataflowBuildConfig(
            folding_config_file="folding.yaml",
            specialize_layers_config_file="specialize_layers.json",
            standalone_thresholds=True,
            max_multithreshold_bit_width=16,
            mvau_wwidth_max=2048,
            verify_steps=[
                # In this custom flow, VerificationStepType is repurposed as follows:
                # Enables "tidied_up_python" (before lowering) verification in step_prepare_graph
                build_cfg.VerificationStepType.TIDY_UP_PYTHON,
                # Enables "lowered_python" (after lowering) and "prepared_graph_python"
                # (after QONNX to FINN conversion) verification in step_prepare_graph
                build_cfg.VerificationStepType.QONNX_TO_FINN_PYTHON,
                # Enables "streamlined_python" verification after custom step_streamline
                build_cfg.VerificationStepType.STREAMLINED_PYTHON,
                # Enables "folded_hls_cppsim" verification after custom step_apply_folding_config
                build_cfg.VerificationStepType.FOLDED_HLS_CPPSIM,
                # Enables RTL simulation of default steps contained in the flow:
                build_cfg.VerificationStepType.NODE_BY_NODE_RTLSIM,  # after step_hw_ipgen
                build_cfg.VerificationStepType.STITCHED_IP_RTLSIM,  # after step_create_stitched_ip
            ],
            # File with test inputs for verification
            verify_input_npy=self._build_inputs["input_npy_path"],
            # File with expected test outputs for verification
            verify_expected_output_npy=self._build_inputs["output_npy_path"],
            # Build steps to execute
            steps=[
                # Prepares the QONNX graph to be consumed by FINN: Cleanup, lowering
                # and Quant to MultiThreshold conversion
                prepare_graph(range_info=range_info),
                # Unified exhaustive streamlining of complex model topologies
                # including attention, residuals and splits
                step_streamline,
                # conversion of the scaled dot-product attention pattern to
                # hardware, including cleanup and data layout squeezing
                step_convert_attention_to_hw,
                # Convert the elementwise binary operations to hardware operators.
                # These include for example adding residual branches and positional encoding
                step_convert_elementwise_binary_to_hw,
                # Convert Lookup layers, e.g., token embedding, to hardware custom operators
                step_convert_lookup_to_hw,
                # Convert Split and Concat operators to hardware, e.g., splits
                # contained in the GLU activation
                step_convert_split_concat_to_hw,
                # Convert depth-wise convolution MatMuls to VVUs
                step_convert_depth_wise_to_hw,
                # Properly replicate the stream feeding the query, key and value projections
                step_replicate_streams,
                # Convert most other layers supported by FINN to HW operators
                "step_convert_to_hw",
                # Specialize HW layer implementations as either HLS or RTL
                "step_specialize_layers",
                "step_create_dataflow_partition",
                # Set the folding configuration to meet the cycles per sequence target
                set_target_parallelization(seq_len, emb_dim),
                # Apply folding configuration, specifying hardware implementation details
                step_apply_folding_config,
                "step_minimize_bit_width",
                # The ScaledDotProductAttention custom op does not define any estimates
                "step_generate_estimate_reports",
                "step_hw_codegen",
                "step_hw_ipgen",
                "step_create_stitched_ip",
                "step_measure_rtlsim_performance",
                "step_out_of_context_synthesis",  # for synthesis results (e.g. utilization)
                "step_synthesize_bitfile",
                "step_make_driver",
                "step_deployment_package",
            ],
        )

        # TESTING custom vs live FIFO-sizing
        if self._params.get("live_fifo_sizing"):
            # insert default FIFO-sizing step (behind step_generate_estimate_reports)
            for i in range(len(cfg.steps)):
                if cfg.steps[i] == "step_generate_estimate_reports":
                    cfg.steps.insert(i + 1, "step_set_fifo_depths")
            # disable verification if live FIFO-sizing is on
            cfg.verify_steps = None
        else:
            # insert Christoph's custom FIFO-sizing step (behind step_hw_ipgen)
            for i in range(len(cfg.steps)):
                if cfg.steps[i] == "step_hw_ipgen":
                    cfg.steps.insert(
                        i + 1, set_fifo_depths(seq_len, emb_dim, uram_threshold=seq_len)
                    )
        return cfg



# fmt: off
# Disable formatter. This is deliberately formatted to stay within 80 characters
# per line. Black, however, formats some lines going beyond this.

# Python builtin math functions: math.ceil returns int, while np.ceil returns
# float
import math

# Numpy math and arrays
import numpy as np

# QONNX/FINN datatypes
from qonnx.core.datatype import DataType

# Multithreshold activations
from qonnx.custom_op.general.multithreshold import multithreshold

# Some utils for working with tensors in qonnx
from qonnx.util.basic import calculate_matvec_accumulator_range

# Derive custom operators form the FINN base custom op
from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp

# FINN logging
from finn.util.logging import log


# Softmax function on numpy arrays with overflow handling matching the HLS
# operator
def softmax(x, axis):
    # For overflow handling, find the maximum value along axis and place ones at
    # each occurrence
    max_ones = (x == np.max(x, axis=axis, keepdims=True)).astype(np.float32)
    # Count the occurrences of the maximum along the normalization axis
    max_counts = np.sum(max_ones, axis=axis, keepdims=True)
    # Exponential of the input
    exp = np.exp(x - np.max(x, axis=axis)[:, np.newaxis])
    # Compute the total along axis
    total = np.sum(exp, axis=axis, keepdims=True)
    # Detect overflow of the summation
    overflow = np.isinf(total)
    # Replace overflows by equal weight given to all instances of the maximum
    # input value. For non overflow just compute normal softmax
    return np.where(overflow, max_ones / max_counts, exp / total)


# Scaled Dot-Product Attention Custom Operator
#   Note: Single head attention
class ScaledDotProductAttention(HWCustomOp):
    # Initializes the operator given an onnx graph node
    def __init__(self, onnx_node, **kwargs):
        # Just forward all arguments to the init method of the CustomOp base
        super().__init__(onnx_node, **kwargs)

    # Node attributes matching the HLS operator
    def get_nodeattr_types(self):
        # Start from parent operator class attributes
        attrs = HWCustomOp.get_nodeattr_types(self)
        # Update attributes dictionary for new custom operator
        attrs.update({
            # Embedding dimension of queries and keys
            "QKDim": ("i", True, 0),
            # Length of the query sequence
            "QLen": ("i", True, 0),
            # Embedding dimension of the values
            "VDim": ("i", True, 0),
            # Length of the key and value sequence
            "KVLen": ("i", True, 0),

            # Folding along the embedding dimensions
            "EmbFold": ("i", True, 0),
            # Folding along the sequence dimensions
            "SeqFold": ("i", True, 0),

            # Datatype of query matrix elements
            "QType": ("s", True, ""),
            # Datatype of key matrix elements
            "KType": ("s", True, ""),
            # Datatype of value matrix elements
            "VType": ("s", True, ""),
            # Datatype of mask matrix elements
            "MType": ("s", False, "INT0"),
            # Datatype of attention weights elements
            "AType": ("s", False, "UINT32"),
            # Datatype of output elements
            "OType": ("s", True, ""),

            # Datatype of accumulator elements of the Query x Key multiplication
            "AccQKMatMul": ("s", False, "UINT32"),
            # Datatype of output elements of the Query x Key multiplication
            "OutQKMatMul": ("s", False, "UINT32"),
            # Activation function type of the Query x Key multiplication
            "ActQKMatMul": ("s", False, "none", {"none", "thresholds"}),
            # Output bias to be applied to the thresholding activation following
            # the Query x Key multiplication
            "BiasActQKMatMul": ("f", False, 0.0),

            # Datatype of accumulator elements of the Attention x Value
            # multiplication
            "AccAVMatMul": ("s", False, "UINT32"),
            # Datatype of output elements of the Attention x Value
            # multiplication
            "OutAVMatMul": ("s", False, "UINT32"),
            # Activation function type of the Attention x Value multiplication
            "ActAVMatMul": ("s", False, "none", {"none", "thresholds"}),
            # Output bias to be applied to the thresholding activation following
            # the Attention x Value multiplication
            "BiasActAVMatMul": ("f", False, 0.0),

            # Scale factor preceding the softmax normalization to dequantize the
            # input
            "DequantSoftmax": ("f", False, 1.0),
            # Datatype of softmax normalization before applying activation or
            # type cast. This is called Acc to stick to the naming scheme of the
            # MatMul operators before.
            #   Note: Currently this is ALWAYS floats
            "AccASoftmax": ("s", False, "FLOAT32"),
            # Activation function type of the softmax normalization of the
            # attention weights
            "ActASoftmax": ("s", False, "none", {"none", "thresholds"}),
            # Output bias to be applied to the thresholding activation following
            # the softmax normalization of the attention weights
            "BiasActASoftmax": ("f", False, 0.0),

            # Mode used for providing the attention mask: There can be no mask,
            # a mask sent as the fourth dynamic input, a mask provided as fourth
            # constant input or a causal attention mask which is generated by
            # the operator itself.
            "mask_mode": (
                "s", True, "none", {"none", "input", "const", "causal"}
            ),

            # Possible execution modes for simulating this node
            #   Note: Override to support python mode
            "exec_mode": (
                "s", False, "python", {"", "rtlsim", "cppsim", "python"}
            ),

            # FPGA resource type for memories/internal buffers of the operator
            # Note: Currently only used for StreamTile buffers
            "ram_style":  (
                "s", False, "auto", {"auto", "block", "distributed", "ultra"}
            ),
            # FPGA resource type for memories of the thresholds parameters
            # Note: Not yet used...
            "ram_style_thresholds": (
                "s", False, "auto", {"auto", "block", "distributed"}
            ),
            # FPGA resource type for memories of the attention mask if the
            # mask_mode is "const"
            "ram_style_mask": (
                "s", False, "auto", {"auto", "block", "distributed"}
            ),
            # FPGA resource type to implement the MAC operations of the two
            # internal matmul operations
            "mac_resource": ("s", False, "auto", {"auto", "lut", "dsp"}),

            # Input and output FIFO depths for multi-I/O nodes
            #   Note: Need to override here as there are three inputs
            "inFIFODepths": ("ints", False, [2, 2, 2]),
            "outFIFODepths": ("ints", False, [2]),
        })
        # Return updated attribute dictionary
        return attrs

    # Shape configuration of the operator
    @property
    def shapes(self):
        # Note: This matches the order of definition above and the order of the
        # HLS lib template as well
        return (self.get_nodeattr("QKDim"), self.get_nodeattr("QLen"),
                self.get_nodeattr("VDim"), self.get_nodeattr("KVLen"))

    # Folding configuration of the operator
    @property
    def folds(self):
        # Note: This matches the order of definition above and the order of the
        # HLS lib template as well
        return self.get_nodeattr("EmbFold"), self.get_nodeattr("SeqFold")

    # Tests whether the given folding is a valid configuration with respect to
    # the shape configuration
    @property
    def is_valid_folding(self):
        # Get and unpack the shape attributes (except the q matrix length, which
        # is never folded)
        qkdim, _, vdim, kvlen = self.shapes
        # Get and unpack the folding attributes
        embfold, seqfold = self.folds
        # All shapes must be multiples of their corresponding fold
        return not ((qkdim % embfold) or (vdim % embfold) or (kvlen % seqfold))

    # Returns an ONNX node that has the same shape inference behavior
    def make_shape_compatible_op(self, model):
        # Infer the output shape from the input shapes
        o_shape = (self.get_nodeattr("QLen"), self.get_nodeattr("VDim"))
        # Get the node wrapped by this custom op
        node = self.onnx_node
        # Get the shape of the input tensor for inferring the number of
        # heads and correctly propagating shapes
        shape = model.get_tensor_shape(node.input[0])
        # Determine the rank of the input tensor to support batched and
        # non-batched inputs
        rank = len(shape)
        # Constant operation producing output of given shape
        #   Note: Rank == 3 allows for propagating yet unrolled multi-attention
        #   heads.
        return super().make_const_shape_op(
            (shape[0], *o_shape) if (rank == 3) else o_shape
        )

    # Infers the output data types and updates the input datatypes of the node
    def infer_node_datatype(self, model):
        # ONNX graph node of the operator
        node = self.onnx_node

        # Get input datatypes from model for query, key, value nodes in order
        q_dtype = model.get_tensor_datatype(node.input[0])
        k_dtype = model.get_tensor_datatype(node.input[1])
        v_dtype = model.get_tensor_datatype(node.input[2])

        # Test for changing query input datatype
        if q_dtype != self.get_nodeattr("QType"):
            # Issue a warning message
            log.warning("QType changing for %s: %s -> %s " % (
                node.name,
                str(self.get_nodeattr("QType")),
                str(q_dtype),
            ))
        # Test for changing key input datatype
        if k_dtype != self.get_nodeattr("KType"):
            # Issue a warning message
            log.warning("KType changing for %s: %s -> %s " % (
                node.name,
                str(self.get_nodeattr("KType")),
                str(k_dtype),
            ))
        # Test for changing value input datatype
        if v_dtype != self.get_nodeattr("VType"):
            # Issue a warning message
            log.warning("VType changing for %s: %s -> %s " % (
                node.name,
                str(self.get_nodeattr("VType")),
                str(v_dtype),
            ))

        # Update the node datatype attributes
        self.set_nodeattr("QType", q_dtype.name)
        self.set_nodeattr("KType", k_dtype.name)
        self.set_nodeattr("VType", v_dtype.name)

        # Attention mask might be provided as an input as well
        if self.get_nodeattr("mask_mode") == "input":
            # Get the datatype attribute of the attention mask
            #   Note: Optional mask will be provided as the fourth input
            mask_dtype = model.get_tensor_datatype(node.input[3])
            # Test for changing mask input datatype
            if mask_dtype != self.get_nodeattr("MType"):
                # Issue a warning message
                log.warning("MType changing for %s: %s -> %s " % (
                    node.name,
                    str(self.get_nodeattr("MType")),
                    str(mask_dtype),
                ))
            # Update the node datatype attribute of the attention mask
            self.set_nodeattr("MType", mask_dtype.name)

        # Set the model output datatype
        model.set_tensor_datatype(
            node.output[0], DataType[self.get_nodeattr('OType')]
        )

    # Executes the attention operator in python mode simulation
    def _execute_node_python(self, context, graph):  # noqa: graph unused
        # Get the node wrapped by this custom op
        node = self.onnx_node

        # Read the input from the execution context and reshape to match the
        # expected folding
        q = context[node.input[0]].reshape(self.get_normal_input_shape(ind=0))
        k = context[node.input[1]].reshape(self.get_normal_input_shape(ind=1))
        v = context[node.input[2]].reshape(self.get_normal_input_shape(ind=2))

        # Quantization activation function following the query and key
        # multiplication
        def act_qk_matmul(x):
            # Only applies if this is specified as a thresholding activation
            if self.get_nodeattr("ActQKMatMul") == "thresholds":
                # Get the thresholds initializer by name from ordered list of
                # optional inputs
                thresholds = context[
                    self.get_input_name_by_name("thresholds_qk_matmul")
                ]
                # Activation value, i.e., bias applied after thresholding
                # activation
                bias = self.get_nodeattr("BiasActQKMatMul")
                # Applies thresholding activation in python to the input
                return multithreshold(x, thresholds) + bias
            # If not thresholds, assume identity function
            return x

        # Quantization activation function following the softmax normalization
        def act_a_softmax(x):
            # Only applies if this is specified as a thresholding activation
            if self.get_nodeattr("ActASoftmax") == "thresholds":
                # Get the thresholds initializer by name from ordered list of
                # optional inputs
                thresholds = context[
                    self.get_input_name_by_name("thresholds_a_softmax")
                ]
                # Activation value, i.e., bias applied after thresholding
                # activation
                bias = self.get_nodeattr("BiasActASoftmax")
                # Applies thresholding activation in python to the input
                return multithreshold(x, thresholds) + bias
            # If not thresholds, assume identity function
            return x

        # Quantization activation function following the attention and values
        # multiplication
        def act_av_matmul(x):
            # Only applies if this is specified as a thresholding activation
            if self.get_nodeattr("ActAVMatMul") == "thresholds":
                # Get the thresholds initializer by name from ordered list of
                # optional inputs
                thresholds = context[
                    self.get_input_name_by_name("thresholds_av_matmul")
                ]
                # Activation value, i.e., bias applied after thresholding
                # activation
                bias = self.get_nodeattr("BiasActAVMatMul")
                # Applies thresholding activation in python to the input
                return multithreshold(x, thresholds) + bias
            # If not thresholds, assume identity function
            return x

        # Scale used to dequantize the qk matrix before computing the softmax in
        # floating point
        dequant = self.get_nodeattr("DequantSoftmax")

        # 1. Queries and keys multiplication followed by quantizing activation
        # function
        qk = act_qk_matmul(np.matmul(q, k.T))

        # Load or create the attention mask for mutually exclusive mask modes

        # There might be no attention mask
        if self.get_nodeattr("mask_mode") == "none":
            # No mask can be realized by adding zero, which does not change
            # anything
            mask = 0
        # There might eb a causal attention mask
        elif self.get_nodeattr("mask_mode") == "causal":
            # A causal mask does not need to be stored and can be generated on
            # the fly
            mask = np.triu(-np.inf * np.ones_like(qk), 1)
        # There might be a constant initializer attention mask
        elif self.get_nodeattr("mask_mode") == "const":
            # Load the mask initializer from the execution context
            mask = context[
                self.get_input_name_by_name("M")
            ]
            # The attention operator represents attention masks as binary masks,
            # but the numpy simulation requires floats with 0 and -inf
            mask = np.where(mask, -np.inf * np.ones_like(mask), 0)
        # The attention mask might be streamed in as the third input
        elif self.get_nodeattr("mask_mode") == "input":
            # Load the mask input from the execution context
            mask = context[
                self.get_input_name_by_name("M")
            ]
            # The attention operator represents attention masks as binary masks,
            # but the numpy simulation requires floats with 0 and -inf
            mask = np.where(mask, -np.inf * np.ones_like(mask), 0)
        # All other mask modes are not supported
        else:
            raise NotImplementedError(
                f"Mask Mode {self.get_nodeattr('mask_mode')} is not implemented"
            )
        # Softmax-normalization of the attention weights followed by quantizing
        # activation function
        a = act_a_softmax(
            # Note: Reshape after masking, as the mask might broadcast messing
            # with the shape
            softmax((dequant * qk + mask).reshape(qk.shape), axis=1)
        )
        # 2. Attention weights and values matmul followed by quantization
        # activation function
        out = act_av_matmul(np.matmul(a, v))

        # Insert the results into the execution context
        context[self.onnx_node.output[0]] = out.reshape(
            self.get_normal_output_shape(ind=0)
        )

    # Executes the attention operator in C++ mode simulation
    def _execute_node_cppsim(self, context, graph):  # noqa: graph unused
        # C++ Simulation needs to be implemented in HLS backend specialization
        raise NotImplementedError(
            f"exec_mode cppsim of {self.__class__.__name__} is not implemented!"
        )

    # Executes the attention operator in RTL mode simulation
    def _execute_node_rtlsim(self, context, graph):  # noqa: graph unused
        # RTL Simulation needs to be implemented in backend specialization
        raise NotImplementedError(
            f"exec_mode rtlsim of {self.__class__.__name__} is not implemented!"
        )

    # Executes the attention operator in simulation (either python, c++ or rtl)
    def execute_node(self, context, graph):
        # Get the configured execution mode
        mode = self.get_nodeattr("exec_mode")
        # Lookup table mapping execution modes to implementing methods
        exec_fns = {
            "python": self._execute_node_python,
            "cppsim": self._execute_node_cppsim,
            "rtlsim": self._execute_node_python,  # TODO: Revert to rtlsim
        }
        # Select and execute the function by mode string
        exec_fns[mode](context, graph)

    # Optional node verification
    def verify_node(self):
        pass

    # Gets the datatype of input at index ind
    def get_input_datatype(self, ind=0):
        # Ordered list of names of allowed inputs
        inputs = ["QType", "KType", "VType"]

        # If the attention mask is provided as input, it has a type as well
        if self.get_nodeattr("mask_mode") == "input":
            # The mask type is an attribute itself
            inputs += ["MType"]

        # TODO: All the following types are probably never requested, they are
        #  implemented for the sake of completeness for now. If they are ever
        #  actually required, check whether the following defaults and dummies
        #  actually still make sense.

        # If there is a thresholding activation for the first matmul, it will
        # have a type as well
        if self.get_nodeattr("ActQKMatMul") == "thresholds":
            # The thresholds will always be of the accumulator type as the
            # activation maps from AccQKMatMul to OutQKMatMul
            inputs += ["AccQKMatMul"]

        # If there is a thresholding activation for the softmax normalization,
        # it will have a type as well
        if self.get_nodeattr("ActASoftmax") == "thresholds":
            # While there is a dummy configurable attribute describing the
            # threshold type of the softmax, these are currently always floats
            inputs += ["AccASoftmax"]

        # If there is a thresholding activation for the second matmul, it will
        # have a type as well
        if self.get_nodeattr("ActAVMatMul") == "thresholds":
            # The thresholds will always be of the accumulator type as the
            # activation maps from AccAVMatMul to OutAVMatMul
            inputs += ["AccAVMatMul"]

        # Look up datatype name in attributes and convert to DataType
        return DataType[self.get_nodeattr(f"{inputs[ind]}")]

    # Gets the datatype of the output (at index ind, but there is just one)
    def get_output_datatype(self, ind=0):
        # Ordered list of names of allowed outputs
        outputs = ["O"]
        # Look up datatype name in attributes and convert to DataType
        return DataType[self.get_nodeattr(f"{outputs[ind]}Type")]

    # Gets the shape of the input at index ind without folding
    def get_normal_input_shape(self, ind=0):
        # List shapes of inputs in order
        inputs_shapes = [
            # Query input sequence
            (self.get_nodeattr("QLen"), self.get_nodeattr("QKDim")),
            # Key input sequence
            (self.get_nodeattr("KVLen"), self.get_nodeattr("QKDim")),
            # Value input sequence
            (self.get_nodeattr("KVLen"), self.get_nodeattr("VDim")),
        ]

        # If the attention mask is provided as input, it has a shape as well
        if self.get_nodeattr("mask_mode") in {"input", "const"}:
            # Mask shape is inferred from query and key sequence lengths
            inputs_shapes += [
                (self.get_nodeattr("QLen"), self.get_nodeattr("KVLen"))
            ]

        # TODO: All the following shapes are probably never requested, they are
        #  implemented for the sake of completeness for now. If they are ever
        #  actually required, remember to insert meaningful shapes.

        # If there is a thresholding activation for the first matmul, these will
        # be the next input index after the (optional) mask
        if self.get_nodeattr("ActQKMatMul") == "thresholds":
            # TODO: This is just a dummy shape
            inputs_shapes += [(0, 0)]

        # If there is a thresholding activation for the softmax normalization,
        # these will be the next (and last) input index after the (optional)
        # second thresholds
        if self.get_nodeattr("ActASoftmax") == "thresholds":
            # TODO: This is just a dummy shape
            inputs_shapes += [(0, 0)]

        # If there is a thresholding activation for the second matmul, these
        # will be the next input index after the (optional) first thresholds
        if self.get_nodeattr("ActAVMatMul") == "thresholds":
            # TODO: This is just a dummy shape
            inputs_shapes += [(0, 0)]

        # Get the shape by indexing into the ordered list of all inputs
        return inputs_shapes[ind]

    # Gets the shape of the output at index ind (there is just one) without
    # folding
    def get_normal_output_shape(self, ind=0):  # noqa, there is just one output
        # The output shape is inferred from the length of the query sequence and
        # the embedding dimension of the values
        return tuple((self.get_nodeattr("QLen"), self.get_nodeattr("VDim")))

    # Gets the shape of the attention weights at index ind (there is just one)
    # without folding
    def get_normal_attention_shape(self, ind=0):  # noqa, there is just one
        # The attention weights have shape covering both sequence dimensions
        return tuple((self.get_nodeattr("QLen"), self.get_nodeattr("KVLen")))

    # Gets the shape of the input at index ind with folding
    def get_folded_input_shape(self, ind=0):
        # Get the unfolded size of the input
        ilen, idim = self.get_normal_input_shape(ind)
        # Get the folding configuration specifying the amount of parallelism
        embfold, seqfold = self.folds

        # Queries, keys and values are all folded similarly along the embedding
        # dimension
        if ind in (0, 1, 2):
            # Note: Embedding dimension is always assumed to be the second
            # dimension, any transpose is handled implicitly by the operator
            return ilen, embfold, idim // embfold

        # If the mask is provided as input, it is folded along the second
        # sequence dimension
        if ind == 3 and self.get_nodeattr("mask_mode") in {"input", "const"}:
            # Note: Both dimensions are sequence dimension, the second
            # corresponds to the KVLen
            return ilen, seqfold, idim // seqfold

        # If this point is reached, probably something went wrong
        # TODO: Requesting the folded shape of thresholds will reach here. Can
        #  this actually happen? Probably it is indeed an error, there should be
        #  no reason to ask for the shape of the thresholds, just ask for the
        #  initializer and get its shape? Folding of the thresholds behaves
        #  differently and would require to actually keep track of mapping
        #  indices to optional inputs to correctly associate the folding
        #  dimensions.
        # TODO: This is just a dummy shape
        return 0, 0, 0

    # Gets the shape of the output at index ind (there is just one) with folding
    def get_folded_output_shape(self, ind=0):  # noqa, there is just one output
        # Get the unfolded size of the output
        olen, odim = self.get_normal_output_shape(ind)
        # Get the folding configuration specifying the amount of parallelism
        embfold, seqfold = self.folds
        # The output is always folded along the embedding dimension, which is
        # assumed to be the second dimension
        return olen, embfold, odim // embfold

    # Gets the shape of the attention weights at index ind (there is just one)
    # with folding
    def get_folded_attention_shape(self, ind=0):  # noqa, there is just one
        # Get the unfolded size of the attention weights
        alen, adim = self.get_normal_attention_shape(ind)
        # Get the folding configuration specifying the amount of parallelism
        embfold, seqfold = self.folds
        # The attention weights are always folded along the sequence dimension,
        # which is assumed to be the second dimension
        return alen, seqfold, adim // seqfold

    # Widths of the input data stream of the input at index ind
    def get_instream_width(self, ind=0):
        # Get the number of bits used to represent the input
        i_bits = self.get_input_datatype(ind).bitwidth()
        # Parallelism is the number of elements in the last dimension of the
        # folded input
        _, _, elems = self.get_folded_input_shape(ind)
        # Width of a stream receiving input elements in parallel
        return elems * i_bits

    # Widths of the output data stream of the output at index ind
    def get_outstream_width(self, ind=0):
        # Get the number of bits used to represent the output
        o_bits = self.get_output_datatype(ind).bitwidth()
        # Parallelism is the number of elements in the last dimension of the
        # folded output
        _, _, elems = self.get_folded_output_shape(ind)
        # Width of a stream producing output elements in parallel
        return elems * o_bits

    # Minimize the accumulator bit width
    def minimize_accumulator_width(self, model):  # noqa: model is unused
        # Get the query, key, value and attention weights type
        QType = DataType[self.get_nodeattr("QType")]  # noqa
        KType = DataType[self.get_nodeattr("KType")]  # noqa
        VType = DataType[self.get_nodeattr("VType")]  # noqa
        AType = DataType[self.get_nodeattr("AType")]  # noqa

        # Compute the worst-case upper and lower bounds of the accumulator range
        lower_worst = QType.min() * np.ones(self.get_normal_input_shape(0))
        lower_range = calculate_matvec_accumulator_range(lower_worst, KType)
        upper_worst = QType.max() * np.ones(self.get_normal_input_shape(0))
        upper_range = calculate_matvec_accumulator_range(  # noqa: Duplicate
            upper_worst, KType
        )
        # Minimum and maximum values of the range
        acc_min = min(min(lower_range), min(upper_range))
        acc_max = max(max(upper_range), max(upper_range))
        # Unsigned accumulator range
        if acc_min >= 0:
            # Number of bits necessary to represent the maximum value of the
            # range. Some values between 0 and acc_min might be unused.
            bitwidth = math.ceil(np.log2(acc_max + 1))
            # New unsigned accumulator datatype of this bitwidth
            AccQKMatMul = DataType[f"UINT{bitwidth}"]  # noqa
        # Signed accumulator range
        else:
            # Maximum absolute value which needs to be represented
            acc_max = max(-acc_min, 1 + acc_max)
            # Number of bits necessary to represent the maximum value of the
            # range. Some values on one of the ends might remain unused.
            bitwidth = math.ceil(np.log2(acc_max) + 1)
            # New signed accumulator datatype of this bitwidth
            AccQKMatMul = DataType[f"INT{bitwidth}"]  # noqa
        # Update the accumulator datatype attribute
        self.set_nodeattr("AccQKMatMul", AccQKMatMul.name)
        # If there is no activation function following the accumulator, the
        # output type needs to be adjusted as well
        if self.get_nodeattr("ActQKMatMul") == "none":
            # Update the output datatype attribute to the same type as the
            # accumulator
            self.set_nodeattr("OutQKMatMul", AccQKMatMul.name)

        # Compute the worst-case upper and lower bounds of the accumulator range
        lower_worst = AType.min() * np.ones(self.get_normal_attention_shape(0))
        lower_range = calculate_matvec_accumulator_range(lower_worst, VType)
        upper_worst = AType.max() * np.ones(self.get_normal_attention_shape(0))
        upper_range = calculate_matvec_accumulator_range(  # noqa: Duplicate
            upper_worst, VType
        )
        # Minimum and maximum values of the range
        acc_min = min(min(lower_range), min(upper_range))
        acc_max = max(max(upper_range), max(upper_range))
        # Unsigned accumulator range
        if acc_min >= 0:
            # Number of bits necessary to represent the maximum value of the
            # range. Some values between 0 and acc_min might be unused.
            bitwidth = math.ceil(np.log2(acc_max + 1))
            # New unsigned accumulator datatype of this bitwidth
            AccAVMatMul = DataType[f"UINT{bitwidth}"]  # noqa
        # Signed accumulator range
        else:
            # Maximum absolute value which needs to be represented
            acc_max = max(-acc_min, 1 + acc_max)
            # Number of bits necessary to represent the maximum value of the
            # range. Some values on one of the ends might remain unused.
            bitwidth = math.ceil(np.log2(acc_max) + 1)
            # New signed accumulator datatype of this bitwidth
            AccAVMatMul = DataType[f"INT{bitwidth}"]  # noqa
        # Update the accumulator datatype attribute
        self.set_nodeattr("AccAVMatMul", AccAVMatMul.name)
        # If there is no activation function following the accumulator, the
        # output type needs to be adjusted as well
        if self.get_nodeattr("ActAVMatMul") == "none":
            # Update the output datatype attribute to the same type as the
            # accumulator
            self.set_nodeattr("OutAVMatMul", AccAVMatMul.name)
            # # The output type of the whole operator is the same as the output
            # # type of the last MatMul
            # TODO: This currently breaks MergeMultiHeads via
            #  MinimizeAccumulatorWidth, which re-infers datatypes after
            #  each custom op instead of once after traversing the whole graph.
            # self.set_nodeattr("OType", AccQKMatMul.name)

    # Gets the number of expected input values, i.e. how many times read()
    # could/should be called on the input stream of this operator
    def get_number_input_values(self, ind=0):
        # Elements over all but the last dimension of the input folded along
        # the embedding dimension
        return np.prod(self.get_folded_input_shape(ind=ind)[:-1])

    # Gets the number of expected output values, i.e. how many times read()
    # could/should be called on the output stream of this operator
    def get_number_output_values(self):
        # Elements over all but the last dimension of the output folded along
        # the embedding dimension
        return np.prod(self.get_folded_output_shape()[:-1])

    # Converts names of optional inputs to the node input index and from there
    # to the ONNX node input name if the input is present.
    #   Note: This mapping is required as the ONNX graph/node may provide
    #   different names (in particular automatically generated unique names) and
    #   some of these are optional inputs.
    def get_input_name_by_name(self, name):
        # Ordered names of the (optional) threshold inputs
        thresholds = [
            "thresholds_qk_matmul",
            "thresholds_a_softmax",
            "thresholds_av_matmul",
        ]

        # Ordered names of primary query, key, value inputs and optional mask
        # and threshold inputs.
        inputs = ["Q", "K", "V", "M", *thresholds]

        # Specify for each input whether it is present or not
        inputs_present = [
            # Note: Primary inputs are always present, the mask is present in
            # "input" or "const" mask mode
            True, True, True, self.get_nodeattr("mask_mode") in {
                "input", "const"
            },
        ]

        # Thresholds are present if the activation function is set to
        # thresholds
        inputs_present.extend([
            self.get_nodeattr("ActQKMatMul") == "thresholds",
            self.get_nodeattr("ActASoftmax") == "thresholds",
            self.get_nodeattr("ActAVMatMul") == "thresholds"
        ])

        # Filter the ordered list of input names for those which are actually
        # present
        inputs = [x for x, present in zip(inputs, inputs_present) if present]

        # Find the position of the requested input name and look up the
        # corresponding input name of the ONNX node
        return self.onnx_node.input[inputs.index(name)]

    # Derives the expected cycles for the attention operation given the folding
    # configuration
    def get_exp_cycles(self):
        # Verify the folding configuration
        assert self.is_valid_folding, \
            f"Invalid folding configuration for {self.onnx_node.name}"
        # Get the input/output dimensions
        qk_dim, q_len, v_dim, kv_len = self.shapes
        # Get folding configuration describing how to parallelize along the
        # dimensions
        emb_fold, seq_fold = self.folds
        # Assume perfect overlap of the constituents of the operator, i.e., of
        # the buffering, both matmul and the softmax, then the expected cycles
        # is the maximum over these operators
        #   Overall worst case cycles without any parallelization: ~ T x T x d
        return max(
            # Transposed keys buffer cycles
            #   Worst case: kv_len * qk_dim, ~ T x d
            kv_len * emb_fold,
            # Queries - keys matmul cycles
            #   Worst case: q_len * qk_dim * kv_len, ~ T x T x d
            q_len * emb_fold * seq_fold,
            # Softmax normalization cycles
            #   Worst case: q_len * kv_len, ~ T x T
            q_len * seq_fold,
            # Values buffer cycles
            #   Worst case: kv_len * v_dim, ~ T x d
            kv_len * emb_fold,
            # Attention weights - values matmul
            #   Worst case: q_len * v_dim * kv_len, ~ T x T x d
            q_len * emb_fold * seq_fold
        )



"""Manage dependencies. Called by run_finn.py"""
from __future__ import annotations

import concurrent
import concurrent.futures
import os
import shlex
import shutil
import subprocess as sp
import sys
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from rich.console import Console
from rich.live import Live
from rich.table import Table
from threading import Lock

from finn.interface import IS_POSIX

FINN_DEPS = {
    "finn-experimental": (
        "https://github.com/Xilinx/finn-experimental.git",
        "0724be21111a21f0d81a072fccc1c446e053f851",
        False,
    ),
    "brevitas": (
        "https://github.com/iksnagreb/brevitas.git",
        "003f9f4070c20639790c7b406a28612a089fc502",
        True,
    ),
    "qonnx": (
        "https://github.com/fpjentzsch/qonnx.git",
        "61863d197f22d893503ba6383ffea8e49b896275",
        True,
    ),
    "dataset_loading": (
        "https://github.com/fbcotter/dataset_loading.git",
        "5b9faa226e5f7c857579d31cdd9acde8cdfb816f",
        True,
    ),
    "cnpy": (
        "https://github.com/rogersce/cnpy.git",
        "4e8810b1a8637695171ed346ce68f6984e585ef4",
        False,
    ),
    "finn-hlslib": (
        "https://github.com/Xilinx/finn-hlslib.git",
        "5dde96382b84979c6caa6f34cdad2ac72fa28489",
        False,
    ),
    "attention-hlslib": (
        "https://github.com/iksnagreb/attention-hlslib.git",
        "afc9720f10e551e1f734e137b21bb6d0a8342177",
        False,
    ),
}

FINN_BOARDFILES = {
    "avnet-bdf": (
        "https://github.com/Avnet/bdf.git",
        "2d49cfc25766f07792c0b314489f21fe916b639b",
        Path(),
    ),
    "xil-bdf": (
        "https://github.com/Xilinx/XilinxBoardStore.git",
        "8cf4bb674a919ac34e3d99d8d71a9e60af93d14e",
        Path("boards/Xilinx/rfsoc2x2"),
    ),
    "rfsoc4x2-bdf": (
        "https://github.com/RealDigitalOrg/RFSoC4x2-BSP.git",
        "13fb6f6c02c7dfd7e4b336b18b959ad5115db696",
        Path("board_files/rfsoc4x2"),
    ),
    "kv260-som-bdf": (
        "https://github.com/Xilinx/XilinxBoardStore.git",
        "98e0d3efc901f0b974006bc4370c2a7ad8856c79",
        Path("boards/Xilinx/kv260_som"),
    ),
    "aupzu3-8gb-bdf": (
        "https://github.com/RealDigitalOrg/aup-zu3-bsp.git",
        "b595ecdf37c7204129517de1773b0895bcdcc2ed",
        Path("board-files/aup-zu3-8gb"),
    ),
}

# URL, do_unzip, where to download to
DIRECT_DOWNLOAD_DEPS = {
    "pynq-z1": (
        "https://github.com/cathalmccabe/pynq-z1_board_files/raw/master/pynq-z1.zip",
        True,
        Path("board_files"),
    ),
    "pynq-z2": (
        "https://dpoauwgwqsy2x.cloudfront.net/Download/pynq-z2.zip",
        True,
        Path("board_files"),
    ),
}

# TODO: Change or make it configurable
GIT_CLONE_TIMEOUT = 120


# Tuple that defines a dep status
# Example: ("qonnx", False, "Wrong commit")
Status = tuple[str, bool, str]


def check_commit(repo: Path, commit: str) -> tuple[bool, str]:
    """Return if the given repo has the correct commit and what commit it read"""
    result = sp.run("git rev-parse HEAD", text=True, capture_output=True, shell=True, cwd=str(repo))
    return result.stdout.strip() == commit, result.stdout.strip()


def run_silent(s: str, loc: str | None | Path, timeout: int | None = None) -> None:
    """Run a command silently directly without shell"""
    sp.run(
        shlex.split(s, posix=IS_POSIX),
        cwd=loc,
        stdout=sp.DEVNULL,
        stderr=sp.DEVNULL,
        stdin=sp.DEVNULL,
        timeout=timeout,
    )


def make_status_table(data: dict[str, tuple[str, str]]) -> Table:
    """Get a dict of the form data[name] = (status, color) and convert it into a table"""
    t = Table()
    t.add_column("Name")
    t.add_column("Status")
    for name, (status, color) in data.items():
        t.add_row(name, f"[bold {color}]{status}[/bold {color}]")
    return t


def update_dependencies(location: Path) -> None:
    """Update dependencies at the given path. Display live status"""
    if not location.exists():
        location.mkdir(parents=True)
    board_file_dir = location / "board_files"
    if not board_file_dir.exists():
        board_file_dir.mkdir(parents=True)
    current_state = {}
    state_lock = Lock()
    any_failed = False
    with Live(make_status_table(current_state)) as live:

        def update_status(key: str, msg: str, color: str) -> None:
            state_lock.acquire()
            current_state[key] = (msg, color)
            live.update(make_status_table(current_state))
            state_lock.release()

        def pull_data(args: tuple) -> bool:
            name, url, do_unzip, target = args
            purl = Path(url)
            target = location.absolute() / target
            update_status(name, "Downloading data...", "orange1")
            if shutil.which("wget") is None:
                update_status(name, "wget not found - could not download data!", "red")
                return False
            if not (target / purl.name).exists():
                result = sp.run(
                    shlex.split(f"wget {url}", posix=IS_POSIX),
                    capture_output=True,
                    text=True,
                    cwd=target,
                )
                if result.returncode != 0:
                    update_status(name, f"Download failed: {result.stderr}", "red")
                    return False
            if do_unzip and not (target / purl.name.replace(purl.suffix, "")).exists():
                update_status(name, "Unzipping data...", "orange1")
                run_silent(f"unzip {purl.name}", target)
            update_status(name, "Dependency ready!", "green")
            return True

        def pull_dep(args: tuple) -> bool:
            pkg_name, giturl, commit, install = args
            target = (location / pkg_name).absolute()
            update_status(pkg_name, "Pulling data...", "orange1")
            if target.exists():
                run_silent("git pull", target)
            else:
                run_silent(f"git clone {giturl} {target}", None, timeout=GIT_CLONE_TIMEOUT)
            if not target.exists():
                update_status(pkg_name, "Bad Git URL or missing network connection", "red")
                return False
            update_status(pkg_name, "Checking out commit...", "orange1")
            run_silent(f"git checkout {commit}", target)
            success, read_commit = check_commit(target, commit)
            if not success:
                update_status(pkg_name, "Failed pulling. Retrying...", "orange1")
                shutil.rmtree(target, ignore_errors=True)
                run_silent(f"git clone {giturl} {target}", None)
                run_silent(f"git checkout {commit}", target)
                success, read_commit = check_commit(target, commit)
            if success:
                if install:
                    update_status(pkg_name, "Installing dependency (pip)!", "orange1")
                    run_silent(f"{sys.executable} -m pip install {target}", None)
                    update_status(pkg_name, "Dependency ready & installed (pip)!", "green")
                else:
                    update_status(pkg_name, "Dependency ready!", "green")
            else:
                update_status(
                    pkg_name,
                    f"Installation failed! Expected commit {commit} but got {read_commit}",
                    "red",
                )
                return False
            return True

        def pull_board(args: tuple) -> bool:
            pkg_name, giturl, commit, copy_from_here = args
            clone_location = location / pkg_name
            copy_source = clone_location / copy_from_here
            copy_target = location / "board_files" / copy_source.name
            update_status(pkg_name, "Pulling data...", "orange1")
            if clone_location.exists():
                run_silent("git pull", clone_location)
            else:
                run_silent(f"git clone {giturl} {clone_location}", None, timeout=GIT_CLONE_TIMEOUT)
            if not clone_location.exists():
                update_status(pkg_name, "Bad Git URL or missing network connection", "red")
                return False
            update_status(pkg_name, "Checking out commit...", "orange1")
            run_silent(f"git checkout {commit}", clone_location)
            success, read_commit = check_commit(clone_location, commit)
            if not success:
                update_status(pkg_name, "Failed pulling. Retrying...", "orange1")
                shutil.rmtree(clone_location, ignore_errors=True)
                run_silent(f"git clone {giturl} {clone_location}", None)
                run_silent(f"git checkout {commit}", clone_location)
                success, read_commit = check_commit(clone_location, commit)
            if success:
                update_status(pkg_name, "Copying boardfiles over...", "orange1")
            else:
                update_status(
                    pkg_name,
                    f"Installation failed! Expected commit {commit} but got {read_commit}",
                    "red",
                )
                return False
            if copy_source != clone_location:
                shutil.copytree(copy_source, copy_target, dirs_exist_ok=True)
            else:
                run_silent(f"cp -r {copy_source}/* {copy_target}", None)
            update_status(pkg_name, "Dependency ready!", "green")
            return True

        with ThreadPoolExecutor(100) as tpe:
            futures = []
            for name, (giturl, commit, install) in FINN_DEPS.items():
                futures.append(tpe.submit(pull_dep, (name, giturl, commit, install)))
            for name, (giturl, commit, copy_from_here) in FINN_BOARDFILES.items():
                futures.append(tpe.submit(pull_board, (name, giturl, commit, copy_from_here)))
            for name, (url, do_unzip, target) in DIRECT_DOWNLOAD_DEPS.items():
                futures.append(tpe.submit(pull_data, (name, url, do_unzip, target)))
            for future in concurrent.futures.as_completed(futures):
                any_failed |= not future.result()

    if any_failed:
        Console().print(
            "[bold red]ERROR: [/bold red][red]"
            "Failed to retrieve all dependencies. Stopping...[/red]"
        )
        sys.exit(1)


def install_finnxsi() -> bool:
    # TODO: integrate properly into the rich.Live above?
    finnxsi_path = os.environ["FINN_XSI"]
    finnxsi_so_path = os.path.join(finnxsi_path, "xsi.so")

    # Set LD_LIBRARY_PATH
    vivado_path = os.environ["XILINX_VIVADO"]
    if "LD_LIBRARY_PATH" not in os.environ.keys():
        os.environ["LD_LIBRARY_PATH"] = f"/lib/x86_64-linux-gnu/:{vivado_path}/lib/lnx64.o"
    else:
        os.environ[
            "LD_LIBRARY_PATH"
        ] = f"/lib/x86_64-linux-gnu/:{vivado_path}/lib/lnx64.o:{os.environ['LD_LIBRARY_PATH']}"

    # Run make
    res = sp.run(["make"], cwd=finnxsi_path, capture_output=True, text=True)
    if res.returncode != 0:
        Console().print(res.stderr)
        return False

    # Check if .so was created
    if not os.path.isfile(finnxsi_so_path):
        return False

    # Set PATH/PYTHONPATH so the .so can be imported
    os.environ["PYTHONPATH"] = f"{os.environ['PYTHONPATH']}:{finnxsi_path}"
    sys.path.append(str(finnxsi_path))
    return True



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# flake8: noqa

call_pynqshell_makefile_template = """
#!/bin/bash
cd %s
export platform=%s
export ip_config=%s
make %s
cd %s
"""

custom_zynq_shell_template = """
set FREQ_MHZ %s
set NUM_AXILITE %d
set NUM_AXIMM %d
set BOARD %s
set FPGA_PART %s
create_project finn_zynq_link ./ -part $FPGA_PART

# Prevent limitation on number of elements for string representations of Vivado collections of objects
# Otherwise we might run into the default limit of 500 if we have many IP_REPO_PATHS
set_param tcl.collectionResultDisplayLimit 0

# set board part repo paths to find PYNQ-Z1/Z2
set paths_prop [get_property BOARD_PART_REPO_PATHS [current_project]]
set paths_param [get_param board.repoPaths]
lappend paths_prop "$BOARDFILES$"
lappend paths_param "$BOARDFILES$"
set_property BOARD_PART_REPO_PATHS $paths_prop [current_project]
set_param board.repoPaths $paths_param

if {$BOARD == "ZCU104"} {
    set_property board_part xilinx.com:zcu104:part0:1.1 [current_project]
    set ZYNQ_TYPE "zynq_us+"
} elseif {$BOARD == "ZCU102"} {
    set_property board_part xilinx.com:zcu102:part0:3.3 [current_project]
    set ZYNQ_TYPE "zynq_us+"
} elseif {$BOARD == "RFSoC2x2"} {
    set_property board_part xilinx.com:rfsoc2x2:part0:1.1 [current_project]
    set ZYNQ_TYPE "zynq_us+"
} elseif {$BOARD == "RFSoC4x2"} {
    set_property board_part realdigital.org:rfsoc4x2:part0:1.0 [current_project]
    set ZYNQ_TYPE "zynq_us+"
} elseif {$BOARD == "Ultra96"} {
    set_property board_part avnet.com:ultra96v1:part0:1.2 [current_project]
    set ZYNQ_TYPE "zynq_us+"
} elseif {$BOARD == "Ultra96-V2"} {
    set_property board_part avnet.com:ultra96v2:part0:1.2 [current_project]
    set ZYNQ_TYPE "zynq_us+"
} elseif {$BOARD == "Pynq-Z2"} {
    set ZYNQ_TYPE "zynq_7000"
    set_property board_part tul.com.tw:pynq-z2:part0:1.0 [current_project]
} elseif {$BOARD == "Pynq-Z1"} {
    set ZYNQ_TYPE "zynq_7000"
    set_property board_part www.digilentinc.com:pynq-z1:part0:1.0 [current_project]
} elseif {$BOARD == "KV260_SOM"} {
    set ZYNQ_TYPE "zynq_us+"
    set_property board_part xilinx.com:kv260_som:part0:1.3 [current_project]
} elseif {$BOARD == "AUP-ZU3_8GB"} {
    set ZYNQ_TYPE "zynq_us+"
    set_property board_part realdigital.org:aup-zu3-8gb:part0:1.0 [current_project]
} else {
    puts "Unrecognized board"
}

create_bd_design "top"
if {$ZYNQ_TYPE == "zynq_us+"} {
    set zynq_ps_vlnv [get_property VLNV [get_ipdefs "xilinx.com:ip:zynq_ultra_ps_e:*"]]
    set zynq_ps_clkname "pl_clk0"
    create_bd_cell -type ip -vlnv $zynq_ps_vlnv zynq_ps
    apply_bd_automation -rule xilinx.com:bd_rule:zynq_ultra_ps_e -config {apply_board_preset "1" }  [get_bd_cells zynq_ps]
    #activate one slave port, deactivate the second master port
    set_property -dict [list CONFIG.PSU__USE__S_AXI_GP2 {1}] [get_bd_cells zynq_ps]
    set_property -dict [list CONFIG.PSU__USE__M_AXI_GP1 {0}] [get_bd_cells zynq_ps]
    #activate one master port and deactivate third master port for AUP-ZU3
    if {$BOARD == "AUP-ZU3_8GB"} {
        set_property -dict [list CONFIG.PSU__USE__M_AXI_GP0 {1}] [get_bd_cells zynq_ps]
        set_property -dict [list CONFIG.PSU__USE__M_AXI_GP2 {0}] [get_bd_cells zynq_ps]
    }
    #set frequency of PS clock (this can't always be exactly met)
    set_property -dict [list CONFIG.PSU__OVERRIDE__BASIC_CLOCK {0}] [get_bd_cells zynq_ps]
    set_property -dict [list CONFIG.PSU__CRL_APB__PL0_REF_CTRL__FREQMHZ [expr int($FREQ_MHZ)]] [get_bd_cells zynq_ps]
} elseif {$ZYNQ_TYPE == "zynq_7000"} {
    set zynq_ps_vlnv [get_property VLNV [get_ipdefs "xilinx.com:ip:processing_system7:*"]]
    set zynq_ps_clkname "FCLK_CLK0"
    create_bd_cell -type ip -vlnv $zynq_ps_vlnv zynq_ps
    apply_bd_automation -rule xilinx.com:bd_rule:processing_system7 -config {make_external "FIXED_IO, DDR" apply_board_preset "1" Master "Disable" Slave "Disable" }  [get_bd_cells zynq_ps]
    set_property -dict [list CONFIG.PCW_USE_S_AXI_HP0 {1}] [get_bd_cells zynq_ps]
    set_property -dict [list CONFIG.PCW_FPGA0_PERIPHERAL_FREQMHZ [expr int($FREQ_MHZ)]] [get_bd_cells zynq_ps]
} else {
    puts "Unrecognized Zynq type"
}

#instantiate axi interconnect, axi smartconnect
set interconnect_vlnv [get_property VLNV [get_ipdefs -all "xilinx.com:ip:axi_interconnect:*" -filter design_tool_contexts=~*IPI*]]
set smartconnect_vlnv [get_property VLNV [get_ipdefs "xilinx.com:ip:smartconnect:*"]]
create_bd_cell -type ip -vlnv $interconnect_vlnv axi_interconnect_0
create_bd_cell -type ip -vlnv $smartconnect_vlnv smartconnect_0
#set number of axilite interfaces, and number of axi master interfaces
set_property -dict [list CONFIG.NUM_SI $NUM_AXIMM] [get_bd_cells smartconnect_0]
set_property -dict [list CONFIG.NUM_MI $NUM_AXILITE] [get_bd_cells axi_interconnect_0]

#create reset controller and connect interconnects to PS
if {$ZYNQ_TYPE == "zynq_us+"} {
    set axi_peripheral_base 0xA0000000
    connect_bd_intf_net [get_bd_intf_pins smartconnect_0/M00_AXI] [get_bd_intf_pins zynq_ps/S_AXI_HP0_FPD]
    connect_bd_intf_net [get_bd_intf_pins zynq_ps/M_AXI_HPM0_FPD] -boundary_type upper [get_bd_intf_pins axi_interconnect_0/S00_AXI]
    #connect interconnect clocks and resets
    apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config { Clk {/zynq_ps/pl_clk0} Freq {} Ref_Clk0 {} Ref_Clk1 {} Ref_Clk2 {}}  [get_bd_pins axi_interconnect_0/ACLK]
    apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config { Clk {/zynq_ps/pl_clk0} Freq {} Ref_Clk0 {} Ref_Clk1 {} Ref_Clk2 {}}  [get_bd_pins axi_interconnect_0/S00_ACLK]
    apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config { Clk {/zynq_ps/pl_clk0} Freq {} Ref_Clk0 {} Ref_Clk1 {} Ref_Clk2 {}}  [get_bd_pins zynq_ps/saxihp0_fpd_aclk]
} elseif {$ZYNQ_TYPE == "zynq_7000"} {
    set axi_peripheral_base 0x40000000
    connect_bd_intf_net -boundary_type upper [get_bd_intf_pins zynq_ps/M_AXI_GP0] [get_bd_intf_pins axi_interconnect_0/S00_AXI]
    connect_bd_intf_net [get_bd_intf_pins smartconnect_0/M00_AXI] [get_bd_intf_pins zynq_ps/S_AXI_HP0]
    apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config { Clk {/zynq_ps/FCLK_CLK0} Freq {} Ref_Clk0 {} Ref_Clk1 {} Ref_Clk2 {}}  [get_bd_pins axi_interconnect_0/ACLK]
    apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config { Clk {/zynq_ps/FCLK_CLK0} Freq {} Ref_Clk0 {} Ref_Clk1 {} Ref_Clk2 {}}  [get_bd_pins axi_interconnect_0/S00_ACLK]
    apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config { Clk {/zynq_ps/FCLK_CLK0} Freq {} Ref_Clk0 {} Ref_Clk1 {} Ref_Clk2 {}}  [get_bd_pins zynq_ps/S_AXI_HP0_ACLK]
}
connect_bd_net [get_bd_pins axi_interconnect_0/ARESETN] [get_bd_pins smartconnect_0/aresetn]

#procedure used by below IP instantiations to map BD address segments based on the axi interface aperture
proc assign_axi_addr_proc {axi_intf_path} {
    #global variable holds current base address
    global axi_peripheral_base
    #infer range
    set range [expr 2**[get_property CONFIG.ADDR_WIDTH [get_bd_intf_pins $axi_intf_path]]]
    set range [expr $range < 4096 ? 4096 : $range]
    #align base address to range
    set offset [expr ($axi_peripheral_base + ($range-1)) & ~($range-1)]
    #perform assignment
    assign_bd_address [get_bd_addr_segs $axi_intf_path/Reg*] -offset $offset -range $range
    #advance base address
    set axi_peripheral_base [expr $offset + $range]
}

#custom IP instantiations/connections start here
%s

# set up debug
if {%d == 1} {
    set_property HDL_ATTRIBUTE.DEBUG true [get_bd_intf_nets {idma0_m_axis_0}]
    set_property HDL_ATTRIBUTE.DEBUG true [get_bd_intf_nets {StreamingDataflowPartition_1_m_axis_0}]
    set_property HDL_ATTRIBUTE.DEBUG true [get_bd_intf_nets {smartconnect_0_M00_AXI}]
    apply_bd_automation -rule xilinx.com:bd_rule:debug -dict [list \
                                                              [get_bd_intf_nets smartconnect_0_M00_AXI] {AXI_R_ADDRESS "Data and Trigger" AXI_R_DATA "Data and Trigger" AXI_W_ADDRESS "Data and Trigger" AXI_W_DATA "Data and Trigger" AXI_W_RESPONSE "Data and Trigger" CLK_SRC "/zynq_ps/FCLK_CLK0" SYSTEM_ILA "Auto" APC_EN "0" } \
                                                              [get_bd_intf_nets idma0_m_axis_0] {AXIS_SIGNALS "Data and Trigger" CLK_SRC "/zynq_ps/FCLK_CLK0" SYSTEM_ILA "Auto" APC_EN "0" } \
                                                              [get_bd_intf_nets StreamingDataflowPartition_1_m_axis_0] {AXIS_SIGNALS "Data and Trigger" CLK_SRC "/zynq_ps/FCLK_CLK0" SYSTEM_ILA "Auto" APC_EN "0" } \
                                                             ]
}

# set up GPIO to trigger reset
if {%d == 1} {
    create_bd_cell -type ip -vlnv xilinx.com:ip:axi_gpio:2.0 axi_gpio_0
    set_property -dict [list CONFIG.C_ALL_OUTPUTS {1} CONFIG.C_DOUT_DEFAULT {0x00000001} CONFIG.C_GPIO_WIDTH {1}] [get_bd_cells axi_gpio_0]
    connect_bd_intf_net [get_bd_intf_pins axi_gpio_0/S_AXI] -boundary_type upper [get_bd_intf_pins axi_interconnect_0/M00_AXI]
    assign_axi_addr_proc axi_gpio_0/S_AXI
    connect_bd_net [get_bd_pins axi_gpio_0/s_axi_aclk] [get_bd_pins axi_interconnect_0/ACLK]
    connect_bd_net [get_bd_pins axi_gpio_0/s_axi_aresetn] [get_bd_pins axi_interconnect_0/ARESETN]
    connect_bd_net [get_bd_pins axi_gpio_0/gpio_io_o] [get_bd_pins rst_zynq_ps_*/aux_reset_in]
}

#finalize clock and reset connections for interconnects
if {$ZYNQ_TYPE == "zynq_us+"} {
    apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config { Clk {/zynq_ps/pl_clk0} }  [get_bd_pins axi_interconnect_0/M*_ACLK]
} elseif {$ZYNQ_TYPE == "zynq_7000"} {
    apply_bd_automation -rule xilinx.com:bd_rule:clkrst -config { Clk {/zynq_ps/FCLK_CLK0} }  [get_bd_pins axi_interconnect_0/M*_ACLK]
}

save_bd_design
assign_bd_address
validate_bd_design

set_property SYNTH_CHECKPOINT_MODE "Hierarchical" [ get_files top.bd ]
make_wrapper -files [get_files top.bd] -import -fileset sources_1 -top

# TODO: make strategies and optimization options configurable
#set_property strategy Flow_PerfOptimized_high [get_runs synth_1]
#set_property STEPS.SYNTH_DESIGN.ARGS.DIRECTIVE AlternateRoutability [get_runs synth_1]
#set_property STEPS.SYNTH_DESIGN.ARGS.RETIMING true [get_runs synth_1]
#set_property strategy Performance_ExtraTimingOpt [get_runs impl_1]
#set_property STEPS.OPT_DESIGN.ARGS.DIRECTIVE Explore [get_runs impl_1]
#set_property STEPS.POST_ROUTE_PHYS_OPT_DESIGN.ARGS.DIRECTIVE AggressiveExplore [get_runs impl_1]
#set_property STEPS.PHYS_OPT_DESIGN.ARGS.DIRECTIVE AggressiveExplore [get_runs impl_1]
#set_property STEPS.POST_ROUTE_PHYS_OPT_DESIGN.IS_ENABLED true [get_runs impl_1]

# out-of-context synth can't be used for bitstream generation
# set_property -name {STEPS.SYNTH_DESIGN.ARGS.MORE OPTIONS} -value {-mode out_of_context} -objects [get_runs synth_1]
# TODO: make number of jobs configurable
launch_runs -jobs 4 -to_step write_bitstream impl_1
wait_on_run [get_runs impl_1]

# generate synthesis report
open_run impl_1
report_utilization -hierarchical -hierarchical_depth 4 -file synth_report.xml -format xml
close_project
"""

vitis_gen_xml_report_tcl_template = """
open_project $VITIS_PROJ_PATH$/_x/link/vivado/vpl/prj/prj.xpr
open_run impl_1
report_utilization -hierarchical -hierarchical_depth 5 -file $VITIS_PROJ_PATH$/synth_report.xml -format xml
"""

# Template scripts for Vivado power estimation
# Initially based on code from Lucas Reuter
# Modified by Felix Jentzsch

template_vivado_open = """
open_project  $PROJ_PATH$
open_run $RUN$
"""

template_vivado_power_fixed = """
#set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -hier -type lut [get_cells -r finn_design_i/.*]
#set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -hier -type register [get_cells -r finn_design_i/.*]
set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -type lut -all
set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -type register -all
set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -type lut_ram -all
set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -type dsp -all
set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -type io_output -all
set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -type bram_enable -all
set_switching_activity -toggle_rate $TOGGLE_RATE$ -static_probability $STATIC_PROB$ -type bram_wr_enable -all

set_switching_activity -deassert_resets
report_power -file $REPORT_PATH$/$REPORT_NAME$.xml -format xml
#reset_switching_activity -hier -type lut [get_cells -r finn_design_i/.*]
#reset_switching_activity -hier -type register [get_cells -r finn_design_i/.*]
"""

template_vivado_power_simulated = """
set_property SOURCE_SET sources_1 [get_filesets sim_1]
import_files -fileset sim_1 -norecurse $TB_FILE_PATH$
set_property top switching_simulation_tb [get_filesets sim_1]
update_compile_order -fileset sim_1

launch_simulation -mode post-implementation -type $SIM_TYPE$
restart
open_saif $SAIF_FILE_PATH$
log_saif [get_objects -r *]
run $SIM_DURATION_NS$ ns
close_saif

read_saif $SAIF_FILE_PATH$
report_power -file $REPORT_PATH$/$REPORT_NAME$.xml -format xml
"""

# TODO: configurable clock frequency instead of hardcoded 100 MHz
template_switching_simulation_tb = """
`timescale 1 ns/10 ps

module switching_simulation_tb;
reg clk;
reg rst;

//dut inputs
reg tready;
reg [$INSTREAM_WIDTH$-1:0] tdata;
reg tvalid;

//dut outputs
wire [$OUTSTREAM_WIDTH$-1:0] accel_tdata;
wire accel_tready;
wire accel_tvalid;

finn_design_wrapper dut(
        .ap_clk(clk),
        .ap_rst_n(rst),
        .m_axis_0_tdata(accel_tdata),
        .m_axis_0_tready(tready),
        .m_axis_0_tvalid(accel_tvalid),
        .s_axis_0_tdata(tdata),
        .s_axis_0_tready(accel_tready),
        .s_axis_0_tvalid(tvalid)
        );

always
    begin
        clk = 0;
        #5;
        clk = 1;
        #5;
    end

integer i;
initial
    begin
        tready = 0;
        tdata = 0;
        tvalid = 0;
        rst = 0;
        #100;
        rst = 1;
        tvalid = 1;
        tready = 1;
        while(1)
            begin
                for (i = 0; i < $INSTREAM_WIDTH$/$DTYPE_WIDTH$; i = i+1) begin
                    tdata[i*$DTYPE_WIDTH$ +: $DTYPE_WIDTH$] = $RANDOM_FUNCTION$;
                end
                #10;
            end
    end
endmodule
"""



# Copyright (C) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnxruntime as rt
from math import ceil
from onnx import TensorProto, helper
from qonnx.core.datatype import DataType
from qonnx.util.basic import qonnx_make_model

from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp
from finn.util.logging import log


class Lookup(HWCustomOp):
    """Abstraction layer for HW implementation of streaming elementwise lookup,
    mapping indices to values."""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {
            # Number of embeddings ("memory depth")
            "NumEmbeddings": ("i", True, 0),
            # Dimensionality of each embedding (part of "memory width")
            "EmbeddingDim": ("i", True, 0),
            # Datatype for embeddings (part of "memory width")
            "EmbeddingType": ("s", True, ""),
            # Datatype for inputs
            "InputType": ("s", True, ""),
            # Input shape
            "InputShape": ("ints", False, [1]),
            # Memory mode
            # internal_embedded : parameters baked into bitfile (BRAM)
            # external : lookup performed in external memory over AXI MM
            "mem_mode": ("s", False, "internal_embedded", ["internal_embedded", "external"]),
            # Width for AXI-MM interface
            # only relevant when mem_mode="external"
            "ext_mem_width": ("i", False, 32),
        }
        my_attrs.update(super().get_nodeattr_types())
        return my_attrs

    def get_exp_cycles(self):
        n_inputs = np.prod(self.get_nodeattr("InputShape"))
        exp_cycles = int(n_inputs)
        return exp_cycles

    def get_normal_input_shape(self, ind=0):
        if ind == 0:
            return self.get_nodeattr("InputShape")
        elif ind == 1:
            return tuple([self.get_nodeattr("NumEmbeddings"), self.get_nodeattr("EmbeddingDim")])
        else:
            raise Exception("Undefined input ind for this layer type")

    def get_normal_output_shape(self, ind=0):
        ishape = self.get_normal_input_shape()
        emb_dim = self.get_nodeattr("EmbeddingDim")
        oshape = list(ishape) + [emb_dim]
        return tuple(oshape)

    def get_folded_input_shape(self, ind=0):
        if ind == 0:
            ishape = self.get_normal_input_shape()
            folded_ishape = list(ishape) + [1]
        else:
            folded_ishape = self.get_normal_input_shape(ind)
        return tuple(folded_ishape)

    def get_folded_output_shape(self, ind=0):
        ishape = self.get_normal_input_shape()
        mem_mode = self.get_nodeattr("mem_mode")
        emb_dim = self.get_nodeattr("EmbeddingDim")
        if mem_mode == "internal_embedded":
            oshape = list(ishape) + [emb_dim]
        elif mem_mode == "external":
            ext_mem_width = self.get_nodeattr("ext_mem_width")
            bits_per_emb_elem = self.get_output_datatype().bitwidth()
            assert ext_mem_width % bits_per_emb_elem == 0
            emb_elems_per_ext_mem_width = ext_mem_width // bits_per_emb_elem
            oshape = list(ishape) + [
                emb_dim // emb_elems_per_ext_mem_width,
                emb_elems_per_ext_mem_width,
            ]
        else:
            raise Exception("Unrecognized mem_mode:" + mem_mode)
        return tuple(oshape)

    def infer_node_datatype(self, model):
        node = self.onnx_node
        idt = model.get_tensor_datatype(node.input[0])
        if idt != self.get_input_datatype():
            warn_str = "InputType changing for %s: %s -> %s " % (
                node.name,
                str(self.get_input_datatype()),
                str(idt),
            )
            log.warning(warn_str)
        self.set_nodeattr("InputType", idt.name)
        odt = DataType[self.get_nodeattr("EmbeddingType")]
        model.set_tensor_datatype(node.output[0], odt)

    def get_input_datatype(self, ind=0):
        if ind == 0:
            ret = DataType[self.get_nodeattr("InputType")]
        elif ind == 1:
            ret = DataType[self.get_nodeattr("EmbeddingType")]
        else:
            raise Exception("Undefined input ind for this layer type")
        return ret

    def get_output_datatype(self, ind=0):
        ret = DataType[self.get_nodeattr("EmbeddingType")]
        return ret

    def get_instream_width(self, ind=0):
        if ind == 0:
            bits = self.get_input_datatype().bitwidth()
        elif ind == 1:
            if self.get_nodeattr("mem_mode") == "internal_embedded":
                bits = 0
            else:
                bits = self.get_nodeattr("ext_mem_width")
        else:
            raise Exception("Undefined input ind for this layer type")
        return bits

    def get_outstream_width(self, ind=0):
        folded_oshape = self.get_folded_output_shape()
        obits = self.get_output_datatype().bitwidth()
        return obits * folded_oshape[-1]

    def execute_node(self, context, graph):
        # create a standard add node to help calculate the result
        node = self.onnx_node
        inp_values = context[node.input[0]]
        ishape = inp_values.shape
        data_values = context[node.input[1]]
        dshape = data_values.shape
        oshape = context[node.output[0]].shape
        inp = helper.make_tensor_value_info(node.input[0], TensorProto.INT64, ishape)
        data = helper.make_tensor_value_info(node.input[1], TensorProto.FLOAT, dshape)
        outp = helper.make_tensor_value_info(node.output[0], TensorProto.FLOAT, oshape)
        node_gather = helper.make_node(
            "Gather",
            inputs=[node.input[1], node.input[0]],
            outputs=[node.output[0]],
        )
        graph_gather = helper.make_graph(
            nodes=[node_gather],
            name="single-gather-exec",
            inputs=[data, inp],
            outputs=[outp],
        )

        opset_version = 13
        opset_imports = [helper.make_opsetid("", opset_version)]
        onnx_kwargs = {"opset_imports": opset_imports}
        model_gather = qonnx_make_model(graph_gather, **onnx_kwargs)
        idict = {node.input[0]: inp_values, node.input[1]: data_values}
        sess = rt.InferenceSession(model_gather.SerializeToString())
        result = sess.run(None, idict)
        context[node.output[0]] = np.asarray(result, dtype=np.float32).reshape(oshape)

    def bram_estimation(self):
        mem_mode = self.get_nodeattr("mem_mode")
        if mem_mode == "internal_embedded":
            # current calculation assumes embeddings always stored in BRAM_18Ks
            # when mem_mode is internal_embedded
            width_factor = ceil(self.get_outstream_width() / 16)
            depth_factor = ceil(self.get_nodeattr("NumEmbeddings") / 1024)
            return width_factor * depth_factor
        else:
            # TODO can we estimate BRAMs for the DMA engine?
            return 0

    def bram_efficiency_estimation(self):
        bram16_est = self.bram_estimation()
        if bram16_est == 0:
            return 1
        ebits = self.get_outstream_width() * self.get_nodeattr("NumEmbeddings")
        bram16_est_capacity = bram16_est * 18 * 1024
        return ebits / bram16_est_capacity

    def get_verilog_top_module_intf_names(self):
        intf_names = super().get_verilog_top_module_intf_names()
        mem_mode = self.get_nodeattr("mem_mode")
        if mem_mode == "external":
            intf_names["axilite"] = ["s_axi_control"]
            intf_names["aximm"] = [("m_axi_gmem", self.get_nodeattr("ext_mem_width"))]
            intf_names["ap_none"] = ["oob_irq"]
        return intf_names
